from datetime import datetime
from ipaddress import IPv4Network
from typing import Any, Dict, Tuple, Union

import pytest

from pycfmodel.model.resources.properties.statement_condition import (
    StatementCondition,
    StatementConditionBuildEvaluatorError,
    build_evaluator,
    build_root_evaluator,
)
from pycfmodel.resolver import resolve


def datetime_in_the_present():
    return datetime(2000, 1, 1, 0, 0)


def datetime_in_the_past():
    return datetime(1990, 1, 1, 0, 0)


def datetime_in_the_future():
    return datetime(2010, 1, 1, 0, 0)


def test_all_possible_conditions():
    # Based on https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_condition_operators.html
    all_operators = set()
    for operator in [
        "ArnEquals",
        "ArnLike",
        "ArnNotEquals",
        "ArnNotLike",
        "Bool",
        "BinaryEquals",
        "DateEquals",
        "DateNotEquals",
        "DateLessThan",
        "DateLessThanEquals",
        "DateGreaterThan",
        "DateGreaterThanEquals",
        "IpAddress",
        "NotIpAddress",
        "NumericEquals",
        "NumericNotEquals",
        "NumericLessThan",
        "NumericLessThanEquals",
        "NumericGreaterThan",
        "NumericGreaterThanEquals",
        "StringEquals",
        "StringNotEquals",
        "StringEqualsIgnoreCase",
        "StringNotEqualsIgnoreCase",
        "StringLike",
        "StringNotLike",
    ]:
        all_operators.add(operator)
        all_operators.add(f"{operator}IfExists")

    # Null
    # Null is not compatible with IfExists
    all_operators.add("Null")

    # For(Any/All)Values
    for operator in all_operators.copy():
        all_operators.add(f"ForAllValues{operator}")
        all_operators.add(f"ForAnyValue{operator}")

    implemented_operators = sorted(StatementCondition.schema()["properties"].keys())
    assert implemented_operators == sorted(all_operators)


def test_statement_condition_remove_colon():
    assert StatementCondition.parse_obj(
        {
            "ForAllValues:ArnEqualsIfExists": {"patata_1": "test_1"},
            "ForAnyValue:ArnEquals": {"patata_2": ["test_2", "test_3"]},
        }
    ) == StatementCondition(
        ForAllValuesArnEqualsIfExists={"patata_1": "test_1"}, ForAnyValueArnEquals={"patata_2": ["test_2", "test_3"]}
    )


def test_eval_is_built_only_if_called():
    condition = StatementCondition()
    assert condition._eval is None
    condition.eval({})
    assert condition._eval is not None


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("Bool", "patata", True, {"patata": True}, True),
        ("Bool", "patata", True, {"patata": False}, False),
        ("Bool", "patata", False, {"patata": True}, False),
        ("Bool", "patata", False, {"patata": False}, True),
    ],
)
def test_build_evaluator_bool(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("IpAddress", "patata", IPv4Network("203.0.113.0/24"), {"patata": IPv4Network("203.0.113.12")}, True),
        ("IpAddress", "patata", IPv4Network("203.0.113.0/24"), {"patata": IPv4Network("10.1.1.1")}, False),
        ("IpAddress", "patata", IPv4Network("203.0.113.10/32"), {"patata": IPv4Network("203.0.113.10")}, True),
        ("IpAddress", "patata", IPv4Network("203.0.113.10/32"), {"patata": IPv4Network("10.1.1.1")}, False),
        ("IpAddress", "patata", IPv4Network("172.16.0.0/12"), {"patata": IPv4Network("203.0.113.10")}, False),
        ("IpAddress", "patata", IPv4Network("172.16.0.0/12"), {"patata": IPv4Network("172.16.0.13")}, True),
        ("IpAddress", "patata", IPv4Network("172.16.0.0/12"), {"patata": IPv4Network("172.16.0.0/12")}, True),
        ("IpAddress", "patata", IPv4Network("172.16.0.0/12"), {"patata": IPv4Network("203.0.113.10/32")}, False),
    ],
)
def test_build_evaluator_ip_address(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("NotIpAddress", "patata", IPv4Network("203.0.113.0/24"), {"patata": IPv4Network("203.0.113.12")}, False),
        ("NotIpAddress", "patata", IPv4Network("203.0.113.0/24"), {"patata": IPv4Network("10.1.1.1")}, True),
        ("NotIpAddress", "patata", IPv4Network("203.0.113.10/32"), {"patata": IPv4Network("203.0.113.10")}, False),
        ("NotIpAddress", "patata", IPv4Network("203.0.113.10/32"), {"patata": IPv4Network("10.1.1.1")}, True),
        ("NotIpAddress", "patata", IPv4Network("172.16.0.0/12"), {"patata": IPv4Network("203.0.113.10")}, True),
        ("NotIpAddress", "patata", IPv4Network("172.16.0.0/12"), {"patata": IPv4Network("172.16.0.13")}, False),
        ("NotIpAddress", "patata", IPv4Network("172.0.0.0/8"), {"patata": IPv4Network("172.0.0.0/8")}, False),
    ],
)
def test_build_evaluator_not_ip_address(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("Null", "patata", True, {"patata": 1}, True),
        ("Null", "patata", True, {}, False),
        ("Null", "patata", False, {"patata": 1}, False),
        ("Null", "patata", False, {}, True),
    ],
)
def test_build_evaluator_null(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("NumericEquals", "patata", 1, {"patata": 1}, True),
        ("NumericEquals", "patata", 1, {"patata": 2}, False),
    ],
)
def test_build_evaluator_numeric_equals(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("NumericNotEquals", "patata", 1, {"patata": 1}, False),
        ("NumericNotEquals", "patata", 1, {"patata": 2}, True),
    ],
)
def test_build_evaluator_numeric_not_equals(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("NumericLessThan", "patata", 1, {"patata": 1}, False),
        ("NumericLessThan", "patata", 1, {"patata": 2}, False),
        ("NumericLessThan", "patata", 2, {"patata": 1}, True),
    ],
)
def test_build_evaluator_numeric_less_than(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("NumericLessThanEquals", "patata", 1, {"patata": 1}, True),
        ("NumericLessThanEquals", "patata", 1, {"patata": 2}, False),
        ("NumericLessThanEquals", "patata", 2, {"patata": 1}, True),
    ],
)
def test_build_evaluator_numeric_less_than_equals(
    function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool
):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("NumericGreaterThan", "patata", 2, {"patata": 2}, False),
        ("NumericGreaterThan", "patata", 2, {"patata": 1}, False),
        ("NumericGreaterThan", "patata", 1, {"patata": 2}, True),
    ],
)
def test_build_evaluator_numeric_greater_than(
    function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool
):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("NumericGreaterThanEquals", "patata", 2, {"patata": 2}, True),
        ("NumericGreaterThanEquals", "patata", 2, {"patata": 1}, False),
        ("NumericGreaterThanEquals", "patata", 1, {"patata": 2}, True),
    ],
)
def test_build_evaluator_numeric_greater_than_equals(
    function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool
):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("DateEquals", "patata", datetime_in_the_present(), {"patata": datetime_in_the_present()}, True),
        ("DateEquals", "patata", datetime_in_the_present(), {"patata": datetime_in_the_past()}, False),
    ],
)
def test_build_evaluator_date_equals(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("DateNotEquals", "patata", datetime_in_the_present(), {"patata": datetime_in_the_present()}, False),
        ("DateNotEquals", "patata", datetime_in_the_present(), {"patata": datetime_in_the_past()}, True),
    ],
)
def test_build_evaluator_date_not_equals(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("DateLessThan", "patata", datetime_in_the_present(), {"patata": datetime_in_the_present()}, False),
        ("DateLessThan", "patata", datetime_in_the_present(), {"patata": datetime_in_the_future()}, False),
        ("DateLessThan", "patata", datetime_in_the_future(), {"patata": datetime_in_the_present()}, True),
    ],
)
def test_build_evaluator_date_less_than(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("DateLessThanEquals", "patata", datetime_in_the_present(), {"patata": datetime_in_the_present()}, True),
        ("DateLessThanEquals", "patata", datetime_in_the_present(), {"patata": datetime_in_the_future()}, False),
        ("DateLessThanEquals", "patata", datetime_in_the_future(), {"patata": datetime_in_the_present()}, True),
    ],
)
def test_build_evaluator_date_less_than_equals(
    function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool
):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("DateGreaterThan", "patata", datetime_in_the_present(), {"patata": datetime_in_the_present()}, False),
        ("DateGreaterThan", "patata", datetime_in_the_future(), {"patata": datetime_in_the_present()}, False),
        ("DateGreaterThan", "patata", datetime_in_the_present(), {"patata": datetime_in_the_future()}, True),
    ],
)
def test_build_evaluator_date_greater_than(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("DateGreaterThanEquals", "patata", datetime_in_the_present(), {"patata": datetime_in_the_present()}, True),
        ("DateGreaterThanEquals", "patata", datetime_in_the_future(), {"patata": datetime_in_the_present()}, False),
        ("DateGreaterThanEquals", "patata", datetime_in_the_present(), {"patata": datetime_in_the_future()}, True),
    ],
)
def test_build_evaluator_date_greater_than_equals(
    function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool
):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("StringEquals", "patata", "aaaaa", {"patata": "aaaaa"}, True),
        ("StringEquals", "patata", "AAAAA", {"patata": "AAAAA"}, True),
        ("StringEquals", "patata", "AaAaA", {"patata": "AaAaA"}, True),
        ("StringEquals", "patata", "AAAAA", {"patata": "aaaaa"}, False),
        ("StringEquals", "patata", "aaaaa", {"patata": "AAAAA"}, False),
        ("StringEquals", "patata", "aAaAa", {"patata": "AaAaA"}, False),
        ("StringEquals", "patata", "aaaaa", {"patata": "bbbbb"}, False),
    ],
)
def test_build_evaluator_string_equals(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("StringNotEquals", "patata", "aaaaa", {"patata": "aaaaa"}, False),
        ("StringNotEquals", "patata", "AAAAA", {"patata": "AAAAA"}, False),
        ("StringNotEquals", "patata", "AaAaA", {"patata": "AaAaA"}, False),
        ("StringNotEquals", "patata", "AAAAA", {"patata": "aaaaa"}, True),
        ("StringNotEquals", "patata", "aaaaa", {"patata": "AAAAA"}, True),
        ("StringNotEquals", "patata", "aAaAa", {"patata": "AaAaA"}, True),
        ("StringNotEquals", "patata", "aaaaa", {"patata": "bbbbb"}, True),
    ],
)
def test_build_evaluator_string_not_equals(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("StringEqualsIgnoreCase", "patata", "aaaaa", {"patata": "aaaaa"}, True),
        ("StringEqualsIgnoreCase", "patata", "AAAAA", {"patata": "AAAAA"}, True),
        ("StringEqualsIgnoreCase", "patata", "AaAaA", {"patata": "AaAaA"}, True),
        ("StringEqualsIgnoreCase", "patata", "AAAAA", {"patata": "aaaaa"}, True),
        ("StringEqualsIgnoreCase", "patata", "aaaaa", {"patata": "AAAAA"}, True),
        ("StringEqualsIgnoreCase", "patata", "aAaAa", {"patata": "AaAaA"}, True),
        ("StringEqualsIgnoreCase", "patata", "aaaaa", {"patata": "bbbbb"}, False),
    ],
)
def test_build_evaluator_string_equals_ignore_case(
    function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool
):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("StringNotEqualsIgnoreCase", "patata", "aaaaa", {"patata": "aaaaa"}, False),
        ("StringNotEqualsIgnoreCase", "patata", "AAAAA", {"patata": "AAAAA"}, False),
        ("StringNotEqualsIgnoreCase", "patata", "AaAaA", {"patata": "AaAaA"}, False),
        ("StringNotEqualsIgnoreCase", "patata", "AAAAA", {"patata": "aaaaa"}, False),
        ("StringNotEqualsIgnoreCase", "patata", "aaaaa", {"patata": "AAAAA"}, False),
        ("StringNotEqualsIgnoreCase", "patata", "aAaAa", {"patata": "AaAaA"}, False),
        ("StringNotEqualsIgnoreCase", "patata", "aaaaa", {"patata": "bbbbb"}, True),
    ],
)
def test_build_evaluator_string_not_equals_ignore_case(
    function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool
):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("StringLike", "patata", "aa*cc", {"patata": "aabcc"}, True),
        ("StringLike", "patata", "aa*cc", {"patata": "abbcc"}, False),
    ],
)
def test_build_evaluator_string_like(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("StringNotLike", "patata", "aa*cc", {"patata": "abbcc"}, True),
        ("StringNotLike", "patata", "aa*cc", {"patata": "aabcc"}, False),
    ],
)
def test_build_evaluator_string_not_like(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        (
            "ArnEquals",
            "patata",
            "arn:aws:ec2:sa-east-1:111122223333:instance/test_instance",
            {"patata": "arn:aws:ec2:sa-east-1:111122223333:instance/test_instance"},
            True,
        ),
        (
            "ArnEquals",
            "patata",
            "arn:aws:ec2:sa-east-1:111122223333:instance/another_test_instance",
            {"patata": "arn:aws:ec2:sa-east-1:111122223333:instance/test_instance"},
            False,
        ),
    ],
)
def test_build_evaluator_arn_equals(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        (
            "ArnNotEquals",
            "patata",
            "arn:aws:ec2:sa-east-1:111122223333:instance/test_instance",
            {"patata": "arn:aws:ec2:sa-east-1:111122223333:instance/test_instance"},
            False,
        ),
        (
            "ArnNotEquals",
            "patata",
            "arn:aws:ec2:sa-east-1:111122223333:instance/another_test_instance",
            {"patata": "arn:aws:ec2:sa-east-1:111122223333:instance/test_instance"},
            True,
        ),
    ],
)
def test_build_evaluator_arn_not_equals(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        (
            "ArnLike",
            "patata",
            "arn:aws:ec2:sa-east-1:111122223333:instance/*",
            {"patata": "arn:aws:ec2:sa-east-1:111122223333:instance/test_instance"},
            True,
        ),
        (
            "ArnLike",
            "patata",
            "arn:aws:ec2:sa-east-1:111122223333:instance/*",
            {"patata": "arn:aws:ec2:eu-west-1:111122223333:instance/test_instance"},
            False,
        ),
    ],
)
def test_build_evaluator_arn_like(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        (
            "ArnNotLike",
            "patata",
            "arn:aws:ec2:sa-east-1:111122223333:instance/*",
            {"patata": "arn:aws:ec2:eu-west-1:111122223333:instance/test_instance"},
            True,
        ),
        (
            "ArnNotLike",
            "patata",
            "arn:aws:ec2:sa-east-1:111122223333:instance/*",
            {"patata": "arn:aws:ec2:sa-east-1:111122223333:instance/test_instance"},
            False,
        ),
    ],
)
def test_build_evaluator_arn_not_like(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arg_a, arg_b, params, expected_output",
    [
        ("BinaryEquals", "patata", bytearray("AaAaA", "utf-8"), {"patata": bytearray("AaAaA", "utf-8")}, True),
        ("BinaryEquals", "patata", bytearray("AAAAA", "utf-8"), {"patata": bytearray("aaaaa", "utf-8")}, False),
    ],
)
def test_build_evaluator_binary_equals(function: str, arg_a: Any, arg_b: Any, params: Dict, expected_output: bool):
    node = build_evaluator(function, arg_a, arg_b)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arguments, params, expected_output",
    [
        ("NumericEquals", {"patata": 1}, {"patata": 1}, True),
        ("NumericEquals", {"patata": 1}, {"patata": 2}, False),
        ("StringEquals", {"patata": ["1", "2", "3"]}, {"patata": ["2", "1"]}, True),
        ("StringEquals", {"patata": ["1", "2", "3"]}, {"patata": ["1", "3"]}, True),
        ("StringEquals", {"patata": "2"}, {"patata": ["2"]}, False),
        ("StringEquals", {"patata": "2"}, {"patata": ["2", "1"]}, False),
        ("StringEquals", {"patata": ["1", "3"]}, {"patata": ["4"]}, False),
        ("StringLike", {"patata": ["sky*"]}, {"patata": ["skyx", "skyscanner", "nope"]}, True),
        ("StringLike", {"patata": ["sky*"]}, {"patata": ["nopeskyx", "nopeskyscanner", "nope"]}, False),
        ("IpAddress", {"patata": [IPv4Network("203.0.113.0/24")]}, {"patata": IPv4Network("203.0.113.0/24")}, True),
        ("IpAddress", {"patata": [IPv4Network("203.0.113.0/24")]}, {"patata": [IPv4Network("203.0.113.0/24")]}, True),
        ("IpAddress", {"patata": ["vpce-123456"]}, {"patata": ["vpce-123456"]}, False),
        ("NotIpAddress", {"patata": ["vpce-123456"]}, {"patata": ["vpce-654321"]}, False),
        (
            "IpAddress",
            {"patata": [IPv4Network("203.0.113.0/24"), IPv4Network("103.0.113.0/24"), IPv4Network("85.0.113.0/24")]},
            {"patata": [IPv4Network("10.0.0.0/8")]},
            False,
        ),
    ],
)
def test_build_root_evaluator(function: str, arguments: Union[Dict, Tuple], params: Dict, expected_output: bool):
    node = build_root_evaluator(function, arguments)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arguments, params, expected_output",
    [
        ("NumericEqualsIfExists", {"patata": 1}, {"patata": 1}, True),
        ("NumericEqualsIfExists", {"patata": 1}, {"patata": 2}, False),
        ("NumericEqualsIfExists", {"patata": 1}, {}, True),
    ],
)
def test_build_root_evaluator_if_exists(
    function: str, arguments: Union[Dict, Tuple], params: Dict, expected_output: bool
):
    node = build_root_evaluator(function, arguments)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arguments, params, expected_output",
    [
        ("ForAllValuesNumericEquals", {"patata": 1}, {"patata": 1}, True),
        ("ForAllValuesNumericEquals", {"patata": [1, 2, 3]}, {"patata": 1}, True),
        ("ForAllValuesNumericEquals", {"patata": 1}, {"patata": [1, 1]}, True),
        ("ForAllValuesNumericEquals", {"patata": [1, 2, 3]}, {"patata": [1, 2]}, True),
        ("ForAllValuesNumericEquals", {"patata": 1}, {"patata": 2}, False),
        ("ForAllValuesNumericEquals", {"patata": [1, 2, 3]}, {"patata": 4}, False),
        ("ForAllValuesNumericEquals", {"patata": 1}, {"patata": [2, 3]}, False),
        ("ForAllValuesNumericEquals", {"patata": [1, 2, 3]}, {"patata": [1, 4]}, False),
    ],
)
def test_build_root_evaluator_for_all_values(
    function: str, arguments: Union[Dict, Tuple], params: Dict, expected_output: bool
):
    node = build_root_evaluator(function, arguments)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arguments, params, expected_output",
    [
        ("ForAllValuesNumericEqualsIfExists", {"patata": 1}, {}, True),
        ("ForAllValuesNumericEqualsIfExists", {"patata": [1, 2, 3]}, {}, True),
        ("ForAllValuesNumericEqualsIfExists", {"patata": 1}, {"patata": 2}, False),
        ("ForAllValuesNumericEqualsIfExists", {"patata": [1, 2, 3]}, {"patata": 4}, False),
        ("ForAllValuesNumericEqualsIfExists", {"patata": 1}, {"patata": [2, 3]}, False),
        ("ForAllValuesNumericEqualsIfExists", {"patata": [1, 2, 3]}, {"patata": [1, 4]}, False),
    ],
)
def test_build_root_evaluator_for_all_values_if_exists(
    function: str, arguments: Union[Dict, Tuple], params: Dict, expected_output: bool
):
    node = build_root_evaluator(function, arguments)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arguments, params, expected_output",
    [
        ("ForAnyValueNumericEquals", {"patata": 1}, {"patata": 1}, True),
        ("ForAnyValueNumericEquals", {"patata": [1, 2, 3]}, {"patata": 1}, True),
        ("ForAnyValueNumericEquals", {"patata": 1}, {"patata": [1, 2]}, True),
        ("ForAnyValueNumericEquals", {"patata": [1, 2, 3]}, {"patata": [1, 2, 5, 6]}, True),
        ("ForAnyValueNumericEquals", {"patata": 1}, {"patata": 2}, False),
        ("ForAnyValueNumericEquals", {"patata": [1, 2, 3]}, {"patata": 4}, False),
        ("ForAnyValueNumericEquals", {"patata": 1}, {"patata": [2, 3]}, False),
        ("ForAnyValueNumericEquals", {"patata": [1, 2, 3]}, {"patata": [4, 5]}, False),
    ],
)
def test_build_root_evaluator_for_any_value(
    function: str, arguments: Union[Dict, Tuple], params: Dict, expected_output: bool
):
    node = build_root_evaluator(function, arguments)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "function, arguments, params, expected_output",
    [
        ("ForAnyValueNumericEqualsIfExists", {"patata": 1}, {}, True),
        ("ForAnyValueNumericEqualsIfExists", {"patata": [1, 2, 3]}, {}, True),
        ("ForAnyValueNumericEqualsIfExists", {"patata": 1}, {"patata": 2}, False),
        ("ForAnyValueNumericEqualsIfExists", {"patata": [1, 2, 3]}, {"patata": 4}, False),
        ("ForAnyValueNumericEqualsIfExists", {"patata": 1}, {"patata": [2, 3]}, False),
        ("ForAnyValueNumericEqualsIfExists", {"patata": [1, 2, 3]}, {"patata": [4, 5, 6]}, False),
    ],
)
def test_build_root_evaluator_for_any_value_if_exists(
    function: str, arguments: Union[Dict, Tuple], params: Dict, expected_output: bool
):
    node = build_root_evaluator(function, arguments)
    assert node(params) == expected_output


@pytest.mark.parametrize(
    "statement_condition, params, expected_output",
    [
        (StatementCondition(), {}, True),
        (StatementCondition(NumericEquals={"patata": 1}), {"patata": 1}, True),
        (StatementCondition(NumericEquals={"patata": 1}), {"patata": 2}, False),
        (
            StatementCondition(NumericEquals={"patata_1": 1}, StringEquals={"patata_2": "A"}),
            {"patata_1": 1, "patata_2": "A"},
            True,
        ),
        (
            StatementCondition(NumericEquals={"patata_1": 1}, StringEquals={"patata_2": "A"}),
            {"patata_1": 2, "patata_2": "A"},
            False,
        ),
        (
            StatementCondition(NumericEquals={"patata_1": 1}, StringEquals={"patata_2": "A"}),
            {"patata_1": 1, "patata_2": "B"},
            False,
        ),
        (
            StatementCondition(NumericEquals={"patata_1": 1}, StringEquals={"patata_2": "A"}),
            {"patata_1": 2, "patata_2": "B"},
            False,
        ),
    ],
)
def test_statement_condition_eval_all_conditions_are_true(
    statement_condition: StatementCondition, params: Dict, expected_output: bool
):
    assert statement_condition.eval(params) == expected_output


def test_statement_condition_without_resolving_raises_error():
    statement_condition_raw = {"StringLike": {"patata": {"Fn::Sub": "${ClusterId}*"}}}
    statement_condition = StatementCondition.parse_obj(statement_condition_raw)
    with pytest.raises(StatementConditionBuildEvaluatorError):
        statement_condition.eval({"patata": "test_cluster"})


def test_statement_condition_with_resolver_works_fine():
    statement_condition_raw = {"StringLike": {"patata": {"Fn::Sub": "${ClusterId}*"}}}
    resolved_statement_condition_raw = resolve(statement_condition_raw, {"ClusterId": "test_cluster"}, {}, {})
    resolved_statement_condition = StatementCondition.parse_obj(resolved_statement_condition_raw)
    assert resolved_statement_condition.eval({"patata": "test_cluster"}) is True



from ipaddress import IPv4Network, IPv6Network
from typing import Optional

from pycfmodel.constants import IPV4_ZERO_VALUE, IPV6_ZERO_VALUE
from pycfmodel.model.resources.properties.property import Property
from pycfmodel.model.types import (
    ResolvableInt,
    ResolvableIntOrStr,
    ResolvableIPv4Network,
    ResolvableIPv6Network,
    ResolvableStr,
)


class SecurityGroupEgressProp(Property):
    """
    An outbound rule permits instances to send traffic from the specified IPv4 or IPv6 CIDR address range, or to the instances associated with the specified security group.

    Properties:

    - CidrIp: The IPv4 ranges.
    - CidrIpv6: The IPv6 ranges.
    - Description: The description of an egress (outbound) security group rule.
    - DestinationPrefixListId: The prefix list IDs for an AWS service.
    - DestinationSecurityGroupId: The ID of the security group.
    - FromPort: The start of port range for the TCP and UDP protocols.
    - IpProtocol: The IP protocol name (tcp, udp, icmp, icmpv6) or number ([see Protocol Numbers](http://www.iana.org/assignments/protocol-numbers/protocol-numbers.xhtml)).
    - ToPort: The end of port range for the TCP and UDP protocols.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-security-group-egress.html)
    """

    CidrIp: Optional[ResolvableIPv4Network] = None
    CidrIpv6: Optional[ResolvableIPv6Network] = None
    Description: Optional[ResolvableStr] = None
    DestinationPrefixListId: Optional[ResolvableStr] = None
    DestinationSecurityGroupId: Optional[ResolvableStr] = None
    FromPort: Optional[ResolvableInt] = None
    IpProtocol: ResolvableIntOrStr
    ToPort: Optional[ResolvableInt] = None

    def ipv4_slash_zero(self) -> bool:
        """Returns True if `CidrIp` matches `0.0.0.0/0`, otherwise False."""
        # Remove after this is fixed https://bugs.python.org/issue38655
        if not self.CidrIp:
            return False
        return self.CidrIp == IPv4Network(IPV4_ZERO_VALUE)

    def ipv6_slash_zero(self) -> bool:
        """Returns True if `CidrIpv6` matches `::/0`, otherwise False."""
        # Remove after this is fixed https://bugs.python.org/issue38655
        if not self.CidrIpv6:
            return False
        return self.CidrIpv6 == IPv6Network(IPV6_ZERO_VALUE)



from typing import ClassVar, Optional

from pycfmodel.model.resources.properties.security_group_ingress_prop import SecurityGroupIngressProp
from pycfmodel.model.resources.resource import Resource
from pycfmodel.model.types import ResolvableStr


class SecurityGroupIngressProperties(SecurityGroupIngressProp):
    """
    Properties:

    - CidrIp: IPv4 address range.
    - CidrIpv6: IPv6 address range.
    - Description: Description for the security group rule.
    - FromPort: Start of port range for the TCP and UDP protocols, or an ICMP/ICMPv6 type number. A value of -1 indicates all ICMP/ICMPv6 types.
    - GroupId: ID of the security group.
    - GroupName: Name of the security group.
    - IpProtocol: IP protocol name.
    - SourcePrefixListId: The prefix list IDs for an AWS service.
    - SourceSecurityGroupId: ID of the security group.
    - SourceSecurityGroupName: Name of the source security group.
    - SourceSecurityGroupOwnerId: AWS account ID for the source security group.
    - ToPort: End of port range for the TCP and UDP protocols, or an ICMP/ICMPv6 code. A value of -1 indicates all ICMP/ICMPv6 codes.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ec2-security-group-ingress.html)
    """

    GroupId: Optional[ResolvableStr] = None
    GroupName: Optional[ResolvableStr] = None


class SecurityGroupIngress(Resource):
    """
    Properties:

    - Properties: A [Security Group Ingress Properties][pycfmodel.model.resources.kms_key.KMSKeyProperties] object.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ec2-security-group-ingress.html)
    """

    TYPE_VALUE: ClassVar = "AWS::EC2::SecurityGroupIngress"
    Type: str = TYPE_VALUE
    Properties: SecurityGroupIngressProperties

    def ipv4_slash_zero(self) -> bool:
        return self.Properties.ipv4_slash_zero()

    def ipv6_slash_zero(self) -> bool:
        return self.Properties.ipv6_slash_zero()



import logging
from base64 import b64encode
from datetime import date
from ipaddress import IPv4Network, IPv6Network
from typing import Dict, List, Union

from pydantic import BaseModel

from pycfmodel.constants import AWS_NOVALUE, CONTAINS_CF_PARAM, CONTAINS_SSM_PARAMETER
from pycfmodel.utils import is_resolvable_dict

logger = logging.getLogger(__file__)

ValidResolvers = Union[str, int, bool, float, List, Dict, date]


def _extended_bool(value) -> bool:
    return _BooleanModel(bool_value=value).bool_value


class _BooleanModel(BaseModel):
    bool_value: bool


def resolve(function: ValidResolvers, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]):
    if function is None:
        return function

    if isinstance(function, str):
        if CONTAINS_SSM_PARAMETER.match(function):
            ssm_parameter_key = CONTAINS_SSM_PARAMETER.match(function).group(1)
            return resolve_ssm(ssm_parameter_key, params)
        return function

    if isinstance(function, (int, float, date, bool, IPv4Network, IPv6Network)):
        return str(function)

    if isinstance(function, list):
        result = []
        for entry in function:
            resolved_value = resolve(entry, params, mappings, conditions)
            if resolved_value != AWS_NOVALUE:
                result.append(resolved_value)
        return result

    if isinstance(function, dict):
        if is_resolvable_dict(function):
            function_name = next(iter(function))
            function_resolver = FUNCTION_MAPPINGS[function_name]
            return function_resolver(function[function_name], params, mappings, conditions)

        result = {}
        for k, v in function.items():
            resolved_value = resolve(v, params, mappings, conditions)
            if resolved_value != AWS_NOVALUE:
                result[k] = resolved_value
        return result

    raise ValueError(f"Not supported type: {type(function)}")


def resolve_ssm(ssm_parameter_key: str, params: Dict) -> str:
    ssm_value = params.get(ssm_parameter_key)
    if not ssm_value or not isinstance(ssm_value, str):
        logger.warning(
            f"Using `UNDEFINED_PARAM_{ssm_parameter_key}` - value not found in AWS SSM or not of string type."
        )
        return f"UNDEFINED_PARAM_{ssm_parameter_key}"
    return ssm_value


def resolve_ref(function_body, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]) -> str:
    resolved_ref = resolve(function_body, params, mappings, conditions)
    if resolved_ref in params:
        return params[resolved_ref]
    else:
        logger.warning(f"Using `UNDEFINED_PARAM_{resolved_ref}` for {resolved_ref}. Original value wasn't available.")
        return f"UNDEFINED_PARAM_{resolved_ref}"


def resolve_join(function_body, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]) -> str:
    delimiter, list_values = function_body
    resolved_delimiter = resolve(delimiter, params, mappings, conditions)
    resolved_list = resolve(list_values, params, mappings, conditions)
    return resolved_delimiter.join(str(e) for e in resolved_list)


def resolve_find_in_map(function_body, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]):
    map_name, top_level_key, second_level_key = function_body
    map_name = resolve(map_name, params, mappings, conditions)
    top_level_key = resolve(top_level_key, params, mappings, conditions)
    second_level_key = resolve(second_level_key, params, mappings, conditions)

    resolved_mapping = mappings.get(map_name, {}).get(top_level_key, {}).get(second_level_key)
    if resolved_mapping:
        return resolved_mapping
    else:
        logger.warning(
            f"Using `UNDEFINED_MAPPING_{map_name}_{top_level_key}_{second_level_key}` for {[map_name, top_level_key, second_level_key]}. Original value wasn't available."
        )
        return f"UNDEFINED_MAPPING_{map_name}_{top_level_key}_{second_level_key}"


def resolve_sub(function_body, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]) -> str:
    replacements = params
    # Whenever we receive a list, first parameter is a text and the second one is a dict with custom replacements.
    # Whenever we receive a string, we need to resolve inlined variables
    if isinstance(function_body, list):
        text, custom_replacements = function_body
        replacements.update(resolve(custom_replacements, params, mappings, conditions))
    else:
        text = function_body
    for match in CONTAINS_CF_PARAM.findall(text):
        match_param = match[2:-1]  # Remove ${ and trailing }
        if match_param in replacements:
            value = resolve(replacements[match_param], params, mappings, conditions)
            text = text.replace(match, str(value))
    return text


def resolve_select(function_body, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]):
    index, list_values = function_body
    resolved_index = int(resolve(index, params, mappings, conditions))
    resolved_list = resolve(list_values, params, mappings, conditions)
    return resolved_list[resolved_index]


def resolve_split(function_body, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]) -> List[str]:
    delimiter, source_string = function_body
    resolved_delimiter = resolve(delimiter, params, mappings, conditions)
    resolved_source_string = resolve(source_string, params, mappings, conditions)
    return resolved_source_string.split(resolved_delimiter)


def resolve_if(function_body, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]) -> str:
    condition, true_section, false_section = function_body
    resolved_funct = resolve({"Condition": condition}, params, mappings, conditions)
    resolved_funct = _extended_bool(resolved_funct)
    if resolved_funct:
        return resolve(true_section, params, mappings, conditions)
    else:
        return resolve(false_section, params, mappings, conditions)


def resolve_and(function_body: List, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]) -> bool:
    return all(_extended_bool(resolve(part, params, mappings, conditions)) for part in function_body)


def resolve_or(function_body: List, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]) -> bool:
    return any(_extended_bool(resolve(part, params, mappings, conditions)) for part in function_body)


def resolve_not(function_body: List, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]) -> bool:
    return not _extended_bool(resolve(function_body[0], params, mappings, conditions))


def resolve_equals(function_body: List, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]) -> bool:
    part_1, part_2 = function_body
    return resolve(part_1, params, mappings, conditions) == resolve(part_2, params, mappings, conditions)


def resolve_base64(function_body: str, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]) -> str:
    resolved_string = resolve(function_body, params, mappings, conditions)
    return str(b64encode(resolved_string.encode("utf-8")), "utf-8")


def resolve_get_attr(function_body, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]) -> str:
    # TODO: Implement.
    logger.warning("`Fn::GetAtt` resolver not implemented, returning `GETATT`")
    return "GETATT"


def resolve_get_azs(function_body, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]) -> str:
    # TODO: Implement.
    logger.warning("`Fn::GetAZs` resolver not implemented, returning `GETAZS`")
    return "GETAZS"


def resolve_condition(function_body, params: Dict, mappings: Dict[str, Dict], conditions: Dict[str, bool]) -> bool:
    # TODO: Implement resolve conditions inside resources
    return conditions.get(function_body, False)


FUNCTION_MAPPINGS = {
    "Condition": resolve_condition,
    "Fn::And": resolve_and,
    "Fn::Base64": resolve_base64,
    "Fn::Equals": resolve_equals,
    "Fn::FindInMap": resolve_find_in_map,
    "Fn::GetAtt": resolve_get_attr,
    "Fn::GetAZs": resolve_get_azs,
    "Fn::If": resolve_if,
    "Fn::ImportValue": resolve_ref,
    "Fn::Join": resolve_join,
    "Fn::Not": resolve_not,
    "Fn::Or": resolve_or,
    "Fn::Select": resolve_select,
    "Fn::Split": resolve_split,
    "Fn::Sub": resolve_sub,
    "Ref": resolve_ref,
}



import pytest

from pycfmodel.model.resources.kms_key import KMSKey


@pytest.fixture()
def kms_key():
    return KMSKey(
        **{
            "Type": "AWS::KMS::Key",
            "Properties": {
                "Description": "Symmetric CMK for signing and verification",
                "KeySpec": "SYMMETRIC_DEFAULT",
                "KeyUsage": "SIGN_VERIFY",
                "MultiRegion": True,
                "EnableKeyRotation": True,
                "KeyPolicy": {
                    "Version": "2012-10-17",
                    "Id": "key-default-1",
                    "Statement": [
                        {
                            "Sid": "Enable IAM User Permissions",
                            "Effect": "Allow",
                            "Principal": {"AWS": "arn:aws:iam::111122223333:root"},
                            "Action": "kms:*",
                            "Resource": "*",
                        },
                        {
                            "Sid": "Allow administration of the key",
                            "Effect": "Allow",
                            "Principal": {"AWS": "arn:aws:iam::111122223333:role/Admin"},
                            "Action": [
                                "kms:Create*",
                                "kms:Describe*",
                                "kms:Enable*",
                                "kms:List*",
                                "kms:Put*",
                                "kms:Update*",
                                "kms:Revoke*",
                                "kms:Disable*",
                                "kms:Get*",
                                "kms:Delete*",
                                "kms:ScheduleKeyDeletion",
                                "kms:CancelKeyDeletion",
                            ],
                            "Resource": "*",
                        },
                        {
                            "Sid": "Allow use of the key",
                            "Effect": "Allow",
                            "Principal": {"AWS": "arn:aws:iam::111122223333:role/Developer"},
                            "Action": ["kms:Sign", "kms:Verify", "kms:DescribeKey"],
                            "Resource": "*",
                        },
                    ],
                },
            },
        }
    )


def test_actions(kms_key):
    assert [
        "kms:CancelKeyDeletion",
        "kms:ConnectCustomKeyStore",
        "kms:CreateAlias",
        "kms:CreateCustomKeyStore",
        "kms:CreateGrant",
        "kms:CreateKey",
        "kms:Decrypt",
        "kms:DeleteAlias",
        "kms:DeleteCustomKeyStore",
        "kms:DeleteImportedKeyMaterial",
        "kms:DescribeCustomKeyStores",
        "kms:DescribeKey",
        "kms:DisableKey",
        "kms:DisableKeyRotation",
        "kms:DisconnectCustomKeyStore",
        "kms:EnableKey",
        "kms:EnableKeyRotation",
        "kms:Encrypt",
        "kms:GenerateDataKey",
        "kms:GenerateDataKeyPair",
        "kms:GenerateDataKeyPairWithoutPlaintext",
        "kms:GenerateDataKeyWithoutPlaintext",
        "kms:GenerateRandom",
        "kms:GetKeyPolicy",
        "kms:GetKeyRotationStatus",
        "kms:GetParametersForImport",
        "kms:GetPublicKey",
        "kms:ImportKeyMaterial",
        "kms:ListAliases",
        "kms:ListGrants",
        "kms:ListKeyPolicies",
        "kms:ListKeys",
        "kms:ListResourceTags",
        "kms:ListRetirableGrants",
        "kms:PutKeyPolicy",
        "kms:ReEncryptFrom",
        "kms:ReEncryptTo",
        "kms:ReplicateKey",
        "kms:RetireGrant",
        "kms:RevokeGrant",
        "kms:ScheduleKeyDeletion",
        "kms:Sign",
        "kms:SynchronizeMultiRegionKey",
        "kms:TagResource",
        "kms:UntagResource",
        "kms:UpdateAlias",
        "kms:UpdateCustomKeyStore",
        "kms:UpdateKeyDescription",
        "kms:UpdatePrimaryRegion",
        "kms:Verify",
    ] == kms_key.Properties.KeyPolicy.get_allowed_actions()


def test_kms_policy_documents(kms_key):
    assert kms_key.policy_documents == []



import re

import pytest

from pycfmodel.model.resources.properties.statement import Statement


def statement_1():
    return Statement(**{"Effect": "Allow", "Action": ["action1"], "NotAction": "action2", "Resource": ["arn"]})


def statement_2():
    return Statement(**{"Effect": "Allow", "Action": "action1", "NotAction": ["action2"], "Resource": ["arn"]})


def statement_3():
    return Statement(**{"Effect": "Allow", "Action": "action1", "Resource": ["arn1"], "NotResource": "arn2"})


def statement_4():
    return Statement(**{"Effect": "Allow", "Action": "action2", "Resource": "arn1", "NotResource": ["arn2"]})


def statement_principal_1():
    return Statement(**{"Principal": {"AWS": "arn:aws:iam::123456789012:root"}})


def statement_principal_2():
    return Statement(
        **{
            "Principal": {
                "AWS": ["arn:aws:iam::AWS-account-ID:user/user-name-1", "arn:aws:iam::AWS-account-ID:user/UserName2"]
            }
        }
    )


def statement_principal_3():
    return Statement(**{"Principal": {"Federated": "cognito-identity.amazonaws.com"}})


def statement_principal_4():
    return Statement(**{"Principal": "arn:aws:iam::123456789012:root"})


def statement_principal_5():
    return Statement(
        **{"Principal": ["arn:aws:iam::AWS-account-ID:user/user-name-1", "arn:aws:iam::AWS-account-ID:user/UserName2"]}
    )


def statement_not_principal_1():
    return Statement(**{"NotPrincipal": {"AWS": "arn:aws:iam::123456789012:root"}})


def statement_not_principal_2():
    return Statement(
        **{
            "NotPrincipal": {
                "AWS": ["arn:aws:iam::AWS-account-ID:user/user-name-1", "arn:aws:iam::AWS-account-ID:user/UserName2"]
            }
        }
    )


def statement_not_principal_3():
    return Statement(**{"NotPrincipal": {"Federated": "cognito-identity.amazonaws.com"}})


def statement_not_principal_4():
    return Statement(**{"NotPrincipal": "arn:aws:iam::123456789012:root"})


def statement_not_principal_5():
    return Statement(
        **{
            "NotPrincipal": [
                "arn:aws:iam::AWS-account-ID:user/user-name-1",
                "arn:aws:iam::AWS-account-ID:user/UserName2",
            ]
        }
    )


def test_capitalize_effect():
    statement = Statement(**{"Effect": "allOw", "Action": ["action1"], "NotAction": "action2", "Resource": ["arn"]})
    assert statement.Effect == "Allow"


@pytest.mark.parametrize(
    "statement, expected_output",
    [
        (statement_1(), ["action1", "action2"]),
        (statement_2(), ["action1", "action2"]),
        (statement_3(), ["action1"]),
        (statement_4(), ["action2"]),
    ],
)
def test_get_action_list(statement, expected_output):
    assert statement.get_action_list() == expected_output


@pytest.mark.parametrize(
    "statement, expected_output",
    [
        (Statement(**{"Effect": "Allow", "Action": "ec2:RunInstances", "Resource": ["arn"]}), ["ec2:RunInstances"]),
        (Statement(**{"Effect": "Allow", "Action": "ec2:Run?nstances", "Resource": ["arn"]}), ["ec2:RunInstances"]),
        (Statement(**{"Effect": "Allow", "Action": "ec?:RunInstances", "Resource": ["arn"]}), ["ec2:RunInstances"]),
        (
            Statement(**{"Effect": "Allow", "Action": "ec2:Run*", "Resource": ["arn"]}),
            ["ec2:RunInstances", "ec2:RunScheduledInstances"],
        ),
    ],
)
def test_get_expanded_action_list(statement, expected_output):
    assert statement.get_expanded_action_list() == expected_output


@pytest.mark.parametrize(
    "statement, expected_output",
    [
        (statement_1(), ["arn"]),
        (statement_2(), ["arn"]),
        (statement_3(), ["arn1", "arn2"]),
        (statement_4(), ["arn1", "arn2"]),
    ],
)
def test_get_resource_list(statement, expected_output):
    assert statement.get_resource_list() == expected_output


@pytest.mark.parametrize(
    "statement, expected_output",
    [
        (statement_principal_1(), ["arn:aws:iam::123456789012:root"]),
        (
            statement_principal_2(),
            ["arn:aws:iam::AWS-account-ID:user/user-name-1", "arn:aws:iam::AWS-account-ID:user/UserName2"],
        ),
        (statement_principal_3(), ["cognito-identity.amazonaws.com"]),
        (statement_principal_4(), ["arn:aws:iam::123456789012:root"]),
        (
            statement_principal_5(),
            ["arn:aws:iam::AWS-account-ID:user/user-name-1", "arn:aws:iam::AWS-account-ID:user/UserName2"],
        ),
        (statement_not_principal_1(), ["arn:aws:iam::123456789012:root"]),
        (
            statement_not_principal_2(),
            ["arn:aws:iam::AWS-account-ID:user/user-name-1", "arn:aws:iam::AWS-account-ID:user/UserName2"],
        ),
        (statement_not_principal_3(), ["cognito-identity.amazonaws.com"]),
        (statement_not_principal_4(), ["arn:aws:iam::123456789012:root"]),
        (
            statement_not_principal_5(),
            ["arn:aws:iam::AWS-account-ID:user/user-name-1", "arn:aws:iam::AWS-account-ID:user/UserName2"],
        ),
    ],
)
def test_get_principal_list(statement, expected_output):
    assert statement.get_principal_list() == expected_output


@pytest.mark.parametrize(
    "statement, pattern, expected_output",
    [
        (statement_1(), re.compile(r"action.*"), ["action1", "action2"]),
        (statement_2(), re.compile(r"action.*"), ["action1", "action2"]),
        (statement_3(), re.compile(r"action2"), []),
        (statement_4(), re.compile(r"a.*2"), ["action2"]),
    ],
)
def test_actions_with(statement, pattern, expected_output):
    assert statement.actions_with(pattern) == expected_output


@pytest.mark.parametrize(
    "statement, pattern, expected_output",
    [
        (statement_principal_1(), re.compile(r"^.*123456789012.*$"), ["arn:aws:iam::123456789012:root"]),
        (statement_principal_2(), re.compile(r"^.*user-name-1$"), ["arn:aws:iam::AWS-account-ID:user/user-name-1"]),
        (statement_principal_3(), re.compile(r"^.*\.amazonaws\.com$"), ["cognito-identity.amazonaws.com"]),
    ],
)
def test_principals_with(statement, pattern, expected_output):
    assert statement.principals_with(pattern) == expected_output


@pytest.mark.parametrize(
    "statement, pattern, expected_output",
    [
        (statement_1(), re.compile(r"^[a]+$"), []),
        (statement_2(), re.compile(r"^[arn]+$"), ["arn"]),
        (statement_3(), re.compile(r"^.*1$"), ["arn1"]),
        (statement_4(), re.compile(r"^.*2$"), ["arn2"]),
    ],
)
def test_resources_with(statement, pattern, expected_output):
    assert statement.resources_with(pattern) == expected_output


@pytest.mark.parametrize(
    "statement, whitelist, expected_output",
    [
        (statement_principal_1(), ["arn:aws:iam::123456789012:root"], []),
        (
            statement_principal_2(),
            ["arn:aws:iam::AWS-account-ID:user/user-name-1"],
            ["arn:aws:iam::AWS-account-ID:user/UserName2"],
        ),
    ],
)
def test_non_whitelisted_principals(statement, whitelist, expected_output):
    assert statement.non_whitelisted_principals(whitelist) == expected_output



import binascii
import logging
from base64 import b64decode
from typing import Any, Callable, Dict, Optional, Tuple, Union
from unicodedata import normalize

from pydantic import root_validator, validator

from pycfmodel.model.base import CustomModel, FunctionDict
from pycfmodel.model.types import (
    ResolvableArnOrList,
    ResolvableBool,
    ResolvableBoolOrList,
    ResolvableBytesOrList,
    ResolvableDatetimeOrList,
    ResolvableIntOrList,
    ResolvableIPOrList,
    ResolvableStrOrList,
)
from pycfmodel.utils import convert_to_list, is_resolvable_dict, not_ip, regex_from_cf_string

logger = logging.getLogger(__name__)


class StatementConditionBuildEvaluatorError(Exception):
    pass


def build_evaluator(function: str, arg_a: Any, arg_b: Any) -> Callable:
    if is_resolvable_dict(arg_b) or isinstance(arg_b, FunctionDict):
        raise StatementConditionBuildEvaluatorError

    if function == "Bool":
        return lambda kwargs: kwargs[arg_a] is arg_b

    elif function == "IpAddress":
        if not_ip(arg_b):
            return lambda _: False
        return lambda kwargs: kwargs[arg_a].subnet_of(arg_b)
    elif function == "NotIpAddress":
        if not_ip(arg_b):
            return lambda _: False
        return lambda kwargs: not kwargs[arg_a].subnet_of(arg_b)

    elif function == "Null":
        return lambda kwargs: (kwargs.get(arg_a) is not None) is arg_b

    elif function in ("StringEquals", "ArnEquals", "BinaryEquals", "NumericEquals", "DateEquals"):
        return lambda kwargs: kwargs[arg_a] == arg_b
    elif function in ("StringNotEquals", "ArnNotEquals", "NumericNotEquals", "DateNotEquals"):
        return lambda kwargs: kwargs[arg_a] != arg_b

    elif function in ("NumericLessThan", "DateLessThan"):
        return lambda kwargs: kwargs[arg_a] < arg_b
    elif function in ("NumericLessThanEquals", "DateLessThanEquals"):
        return lambda kwargs: kwargs[arg_a] <= arg_b
    elif function in ("NumericGreaterThan", "DateGreaterThan"):
        return lambda kwargs: kwargs[arg_a] > arg_b
    elif function in ("NumericGreaterThanEquals", "DateGreaterThanEquals"):
        return lambda kwargs: kwargs[arg_a] >= arg_b

    elif function == "StringEqualsIgnoreCase":
        arg_b = normalize("NFKD", arg_b.casefold())
        return lambda kwargs: normalize("NFKD", kwargs[arg_a].casefold()) == arg_b
    elif function == "StringNotEqualsIgnoreCase":
        arg_b = normalize("NFKD", arg_b.casefold())
        return lambda kwargs: normalize("NFKD", kwargs[arg_a].casefold()) != arg_b
    elif function in ("StringLike", "ArnLike"):
        arg_b = regex_from_cf_string(arg_b)
        return lambda kwargs: bool(arg_b.match(kwargs[arg_a]))
    elif function in ("StringNotLike", "ArnNotLike"):
        arg_b = regex_from_cf_string(arg_b)
        return lambda kwargs: not bool(arg_b.match(kwargs[arg_a]))
    else:
        logger.error(f"{function} is not supported.")
        raise NotImplementedError


def build_root_evaluator(function: str, arguments: Union[Dict, Tuple]) -> Callable:
    if isinstance(arguments, dict):
        arguments = arguments.items()
    elif isinstance(arguments, tuple):
        arguments = [arguments]

    if function.endswith("IfExists"):
        new_function = function.replace("IfExists", "")
    elif function.startswith("ForAllValues"):
        new_function = function.replace("ForAllValues", "")
    elif function.startswith("ForAnyValue"):
        new_function = function.replace("ForAnyValue", "")
    else:
        new_function = function

    group_of_nodes = []
    for arg_a, arg_b in arguments:
        if function.endswith("IfExists"):
            node = build_root_evaluator(new_function, (arg_a, arg_b))
            group_of_nodes.append(lambda kwargs: node(kwargs) if kwargs.get(arg_a) is not None else True)
        elif function.startswith("ForAllValues"):
            nodes = [build_root_evaluator(new_function, (arg_a, item)) for item in convert_to_list(arg_b)]
            all_nodes = lambda kwargs: any(node(kwargs) for node in nodes)  # noqa: E731
            group_of_nodes.append(
                lambda kwargs: all(all_nodes({**kwargs, arg_a: item}) for item in convert_to_list(kwargs[arg_a]))
            )
        elif function.startswith("ForAnyValue") or isinstance(arg_b, list):
            nodes = [build_root_evaluator(new_function, (arg_a, item)) for item in convert_to_list(arg_b)]
            all_nodes = lambda kwargs: any(node(kwargs) for node in nodes)  # noqa: E731
            group_of_nodes.append(
                lambda kwargs: any(all_nodes({**kwargs, arg_a: item}) for item in convert_to_list(kwargs[arg_a]))
            )
        else:
            group_of_nodes.append(build_evaluator(new_function, arg_a, arg_b))

    return lambda kwargs: all(group(kwargs) for group in group_of_nodes)


class StatementCondition(CustomModel):
    """
    Contains the condition to be matched to apply the statement that belongs to.

    | Type                       | Operators                 | ...IfExists | ForAllValues... | ForAnyValue... |
    | -------------------------- | ------------------------- | ----------- | --------------- | -------------- |
    | String                     | StringEquals              | Yes         | Yes             | Yes            |
    | String                     | StringNotEquals           | Yes         | Yes             | Yes            |
    | String                     | StringEqualsIgnoreCase    | Yes         | Yes             | Yes            |
    | String                     | StringNotEqualsIgnoreCase | Yes         | Yes             | Yes            |
    | String                     | StringLike                | Yes         | Yes             | Yes            |
    | String                     | StringNotLike             | Yes         | Yes             | Yes            |
    | Numeric                    | NumericEquals             | Yes         | Yes             | Yes            |
    | Numeric                    | NumericNotEquals          | Yes         | Yes             | Yes            |
    | Numeric                    | NumericLessThan           | Yes         | Yes             | Yes            |
    | Numeric                    | NumericLessThanEquals     | Yes         | Yes             | Yes            |
    | Numeric                    | NumericGreaterThan        | Yes         | Yes             | Yes            |
    | Numeric                    | NumericGreaterThanEquals  | Yes         | Yes             | Yes            |
    | Date and time              | DateEquals                | Yes         | Yes             | Yes            |
    | Date and time              | DateNotEquals             | Yes         | Yes             | Yes            |
    | Date and time              | DateLessThan              | Yes         | Yes             | Yes            |
    | Date and time              | DateLessThanEquals        | Yes         | Yes             | Yes            |
    | Date and time              | DateGreaterThan           | Yes         | Yes             | Yes            |
    | Date and time              | DateGreaterThanEquals     | Yes         | Yes             | Yes            |
    | Boolean                    | Bool                      | Yes         | Yes             | Yes            |
    | Binary                     | BinaryEquals              | Yes         | Yes             | Yes            |
    | IP address                 | IpAddress                 | Yes         | Yes             | Yes            |
    | IP address                 | NotIpAddress              | Yes         | Yes             | Yes            |
    | Amazon Resource Name (ARN) | ArnEquals                 | Yes         | Yes             | Yes            |
    | Amazon Resource Name (ARN) | ArnLike                   | Yes         | Yes             | Yes            |
    | Amazon Resource Name (ARN) | ArnNotEquals              | Yes         | Yes             | Yes            |
    | Amazon Resource Name (ARN) | ArnNotLike                | Yes         | Yes             | Yes            |
    | Existence                  | Null                      | No          | Yes             | Yes            |

    Table based on [AWS Docs](https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements_condition_operators.html)

    For conditions such as `StringEquals` with multiple values for one key, we evaluate them using the logical `OR`,
    similar to if the condition key was `ForAnyValue:StringEquals`. This follows the
    [documentation from AWS](https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_multi-value-conditions.html).
    """

    ArnEquals: Optional[Dict[str, ResolvableArnOrList]]
    ArnLike: Optional[Dict[str, ResolvableArnOrList]]
    ArnNotEquals: Optional[Dict[str, ResolvableArnOrList]]
    ArnNotLike: Optional[Dict[str, ResolvableArnOrList]]
    ArnEqualsIfExists: Optional[Dict[str, ResolvableArnOrList]]
    ArnLikeIfExists: Optional[Dict[str, ResolvableArnOrList]]
    ArnNotEqualsIfExists: Optional[Dict[str, ResolvableArnOrList]]
    ArnNotLikeIfExists: Optional[Dict[str, ResolvableArnOrList]]
    ForAllValuesArnEquals: Optional[Dict[str, ResolvableArnOrList]]
    ForAllValuesArnLike: Optional[Dict[str, ResolvableArnOrList]]
    ForAllValuesArnNotEquals: Optional[Dict[str, ResolvableArnOrList]]
    ForAllValuesArnNotLike: Optional[Dict[str, ResolvableArnOrList]]
    ForAllValuesArnEqualsIfExists: Optional[Dict[str, ResolvableArnOrList]]
    ForAllValuesArnLikeIfExists: Optional[Dict[str, ResolvableArnOrList]]
    ForAllValuesArnNotEqualsIfExists: Optional[Dict[str, ResolvableArnOrList]]
    ForAllValuesArnNotLikeIfExists: Optional[Dict[str, ResolvableArnOrList]]
    ForAnyValueArnEquals: Optional[Dict[str, ResolvableArnOrList]]
    ForAnyValueArnLike: Optional[Dict[str, ResolvableArnOrList]]
    ForAnyValueArnNotEquals: Optional[Dict[str, ResolvableArnOrList]]
    ForAnyValueArnNotLike: Optional[Dict[str, ResolvableArnOrList]]
    ForAnyValueArnEqualsIfExists: Optional[Dict[str, ResolvableArnOrList]]
    ForAnyValueArnLikeIfExists: Optional[Dict[str, ResolvableArnOrList]]
    ForAnyValueArnNotEqualsIfExists: Optional[Dict[str, ResolvableArnOrList]]
    ForAnyValueArnNotLikeIfExists: Optional[Dict[str, ResolvableArnOrList]]

    Bool: Optional[Dict[str, ResolvableBoolOrList]]
    BoolIfExists: Optional[Dict[str, ResolvableBoolOrList]]
    ForAllValuesBool: Optional[Dict[str, ResolvableBoolOrList]]
    ForAllValuesBoolIfExists: Optional[Dict[str, ResolvableBoolOrList]]
    ForAnyValueBool: Optional[Dict[str, ResolvableBoolOrList]]
    ForAnyValueBoolIfExists: Optional[Dict[str, ResolvableBoolOrList]]

    BinaryEquals: Optional[Dict[str, ResolvableBytesOrList]]
    BinaryEqualsIfExists: Optional[Dict[str, ResolvableBytesOrList]]
    ForAllValuesBinaryEquals: Optional[Dict[str, ResolvableBytesOrList]]
    ForAllValuesBinaryEqualsIfExists: Optional[Dict[str, ResolvableBytesOrList]]
    ForAnyValueBinaryEquals: Optional[Dict[str, ResolvableBytesOrList]]
    ForAnyValueBinaryEqualsIfExists: Optional[Dict[str, ResolvableBytesOrList]]

    DateEquals: Optional[Dict[str, ResolvableDatetimeOrList]]
    DateNotEquals: Optional[Dict[str, ResolvableDatetimeOrList]]
    DateLessThan: Optional[Dict[str, ResolvableDatetimeOrList]]
    DateLessThanEquals: Optional[Dict[str, ResolvableDatetimeOrList]]
    DateGreaterThan: Optional[Dict[str, ResolvableDatetimeOrList]]
    DateGreaterThanEquals: Optional[Dict[str, ResolvableDatetimeOrList]]
    DateEqualsIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    DateNotEqualsIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    DateLessThanIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    DateLessThanEqualsIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    DateGreaterThanIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    DateGreaterThanEqualsIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAllValuesDateEquals: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAllValuesDateNotEquals: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAllValuesDateLessThan: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAllValuesDateLessThanEquals: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAllValuesDateGreaterThan: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAllValuesDateGreaterThanEquals: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAllValuesDateEqualsIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAllValuesDateNotEqualsIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAllValuesDateLessThanIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAllValuesDateLessThanEqualsIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAllValuesDateGreaterThanIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAllValuesDateGreaterThanEqualsIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAnyValueDateEquals: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAnyValueDateNotEquals: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAnyValueDateLessThan: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAnyValueDateLessThanEquals: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAnyValueDateGreaterThan: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAnyValueDateGreaterThanEquals: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAnyValueDateEqualsIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAnyValueDateNotEqualsIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAnyValueDateLessThanIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAnyValueDateLessThanEqualsIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAnyValueDateGreaterThanIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]
    ForAnyValueDateGreaterThanEqualsIfExists: Optional[Dict[str, ResolvableDatetimeOrList]]

    IpAddress: Optional[Dict[str, Union[ResolvableIPOrList, ResolvableStrOrList]]]
    NotIpAddress: Optional[Dict[str, Union[ResolvableIPOrList, ResolvableStrOrList]]]
    IpAddressIfExists: Optional[Dict[str, Union[ResolvableIPOrList, ResolvableStrOrList]]]
    NotIpAddressIfExists: Optional[Dict[str, Union[ResolvableIPOrList, ResolvableStrOrList]]]
    ForAllValuesIpAddress: Optional[Dict[str, Union[ResolvableIPOrList, ResolvableStrOrList]]]
    ForAllValuesNotIpAddress: Optional[Dict[str, Union[ResolvableIPOrList, ResolvableStrOrList]]]
    ForAnyValueIpAddress: Optional[Dict[str, Union[ResolvableIPOrList, ResolvableStrOrList]]]
    ForAnyValueNotIpAddress: Optional[Dict[str, Union[ResolvableIPOrList, ResolvableStrOrList]]]
    ForAllValuesIpAddressIfExists: Optional[Dict[str, Union[ResolvableIPOrList, ResolvableStrOrList]]]
    ForAllValuesNotIpAddressIfExists: Optional[Dict[str, Union[ResolvableIPOrList, ResolvableStrOrList]]]
    ForAnyValueIpAddressIfExists: Optional[Dict[str, Union[ResolvableIPOrList, ResolvableStrOrList]]]
    ForAnyValueNotIpAddressIfExists: Optional[Dict[str, Union[ResolvableIPOrList, ResolvableStrOrList]]]

    Null: Optional[Dict[str, ResolvableBool]]
    ForAllValuesNull: Optional[Dict[str, ResolvableBool]]
    ForAnyValueNull: Optional[Dict[str, ResolvableBool]]

    NumericEquals: Optional[Dict[str, ResolvableIntOrList]]
    NumericNotEquals: Optional[Dict[str, ResolvableIntOrList]]
    NumericLessThan: Optional[Dict[str, ResolvableIntOrList]]
    NumericLessThanEquals: Optional[Dict[str, ResolvableIntOrList]]
    NumericGreaterThan: Optional[Dict[str, ResolvableIntOrList]]
    NumericGreaterThanEquals: Optional[Dict[str, ResolvableIntOrList]]
    NumericEqualsIfExists: Optional[Dict[str, ResolvableIntOrList]]
    NumericNotEqualsIfExists: Optional[Dict[str, ResolvableIntOrList]]
    NumericLessThanIfExists: Optional[Dict[str, ResolvableIntOrList]]
    NumericLessThanEqualsIfExists: Optional[Dict[str, ResolvableIntOrList]]
    NumericGreaterThanIfExists: Optional[Dict[str, ResolvableIntOrList]]
    NumericGreaterThanEqualsIfExists: Optional[Dict[str, ResolvableIntOrList]]
    ForAllValuesNumericEquals: Optional[Dict[str, ResolvableIntOrList]]
    ForAllValuesNumericNotEquals: Optional[Dict[str, ResolvableIntOrList]]
    ForAllValuesNumericLessThan: Optional[Dict[str, ResolvableIntOrList]]
    ForAllValuesNumericLessThanEquals: Optional[Dict[str, ResolvableIntOrList]]
    ForAllValuesNumericGreaterThan: Optional[Dict[str, ResolvableIntOrList]]
    ForAllValuesNumericGreaterThanEquals: Optional[Dict[str, ResolvableIntOrList]]
    ForAllValuesNumericEqualsIfExists: Optional[Dict[str, ResolvableIntOrList]]
    ForAllValuesNumericNotEqualsIfExists: Optional[Dict[str, ResolvableIntOrList]]
    ForAllValuesNumericLessThanIfExists: Optional[Dict[str, ResolvableIntOrList]]
    ForAllValuesNumericLessThanEqualsIfExists: Optional[Dict[str, ResolvableIntOrList]]
    ForAllValuesNumericGreaterThanIfExists: Optional[Dict[str, ResolvableIntOrList]]
    ForAllValuesNumericGreaterThanEqualsIfExists: Optional[Dict[str, ResolvableIntOrList]]
    ForAnyValueNumericEquals: Optional[Dict[str, ResolvableIntOrList]]
    ForAnyValueNumericNotEquals: Optional[Dict[str, ResolvableIntOrList]]
    ForAnyValueNumericLessThan: Optional[Dict[str, ResolvableIntOrList]]
    ForAnyValueNumericLessThanEquals: Optional[Dict[str, ResolvableIntOrList]]
    ForAnyValueNumericGreaterThan: Optional[Dict[str, ResolvableIntOrList]]
    ForAnyValueNumericGreaterThanEquals: Optional[Dict[str, ResolvableIntOrList]]
    ForAnyValueNumericEqualsIfExists: Optional[Dict[str, ResolvableIntOrList]]
    ForAnyValueNumericNotEqualsIfExists: Optional[Dict[str, ResolvableIntOrList]]
    ForAnyValueNumericLessThanIfExists: Optional[Dict[str, ResolvableIntOrList]]
    ForAnyValueNumericLessThanEqualsIfExists: Optional[Dict[str, ResolvableIntOrList]]
    ForAnyValueNumericGreaterThanIfExists: Optional[Dict[str, ResolvableIntOrList]]
    ForAnyValueNumericGreaterThanEqualsIfExists: Optional[Dict[str, ResolvableIntOrList]]

    StringEquals: Optional[Dict[str, ResolvableStrOrList]]
    StringNotEquals: Optional[Dict[str, ResolvableStrOrList]]
    StringEqualsIgnoreCase: Optional[Dict[str, ResolvableStrOrList]]
    StringNotEqualsIgnoreCase: Optional[Dict[str, ResolvableStrOrList]]
    StringLike: Optional[Dict[str, ResolvableStrOrList]]
    StringNotLike: Optional[Dict[str, ResolvableStrOrList]]
    StringEqualsIfExists: Optional[Dict[str, ResolvableStrOrList]]
    StringNotEqualsIfExists: Optional[Dict[str, ResolvableStrOrList]]
    StringEqualsIgnoreCaseIfExists: Optional[Dict[str, ResolvableStrOrList]]
    StringNotEqualsIgnoreCaseIfExists: Optional[Dict[str, ResolvableStrOrList]]
    StringLikeIfExists: Optional[Dict[str, ResolvableStrOrList]]
    StringNotLikeIfExists: Optional[Dict[str, ResolvableStrOrList]]
    ForAllValuesStringEquals: Optional[Dict[str, ResolvableStrOrList]]
    ForAllValuesStringNotEquals: Optional[Dict[str, ResolvableStrOrList]]
    ForAllValuesStringEqualsIgnoreCase: Optional[Dict[str, ResolvableStrOrList]]
    ForAllValuesStringNotEqualsIgnoreCase: Optional[Dict[str, ResolvableStrOrList]]
    ForAllValuesStringLike: Optional[Dict[str, ResolvableStrOrList]]
    ForAllValuesStringNotLike: Optional[Dict[str, ResolvableStrOrList]]
    ForAllValuesStringEqualsIfExists: Optional[Dict[str, ResolvableStrOrList]]
    ForAllValuesStringNotEqualsIfExists: Optional[Dict[str, ResolvableStrOrList]]
    ForAllValuesStringEqualsIgnoreCaseIfExists: Optional[Dict[str, ResolvableStrOrList]]
    ForAllValuesStringNotEqualsIgnoreCaseIfExists: Optional[Dict[str, ResolvableStrOrList]]
    ForAllValuesStringLikeIfExists: Optional[Dict[str, ResolvableStrOrList]]
    ForAllValuesStringNotLikeIfExists: Optional[Dict[str, ResolvableStrOrList]]
    ForAnyValueStringEquals: Optional[Dict[str, ResolvableStrOrList]]
    ForAnyValueStringNotEquals: Optional[Dict[str, ResolvableStrOrList]]
    ForAnyValueStringEqualsIgnoreCase: Optional[Dict[str, ResolvableStrOrList]]
    ForAnyValueStringNotEqualsIgnoreCase: Optional[Dict[str, ResolvableStrOrList]]
    ForAnyValueStringLike: Optional[Dict[str, ResolvableStrOrList]]
    ForAnyValueStringNotLike: Optional[Dict[str, ResolvableStrOrList]]
    ForAnyValueStringEqualsIfExists: Optional[Dict[str, ResolvableStrOrList]]
    ForAnyValueStringNotEqualsIfExists: Optional[Dict[str, ResolvableStrOrList]]
    ForAnyValueStringEqualsIgnoreCaseIfExists: Optional[Dict[str, ResolvableStrOrList]]
    ForAnyValueStringNotEqualsIgnoreCaseIfExists: Optional[Dict[str, ResolvableStrOrList]]
    ForAnyValueStringLikeIfExists: Optional[Dict[str, ResolvableStrOrList]]
    ForAnyValueStringNotLikeIfExists: Optional[Dict[str, ResolvableStrOrList]]

    _eval: Optional[Callable] = None

    @root_validator(pre=True)
    def remove_colon(cls, values):
        return {key.replace(":", ""): value for key, value in values.items()}

    @validator(
        "BinaryEquals",
        "BinaryEqualsIfExists",
        "ForAllValuesBinaryEquals",
        "ForAllValuesBinaryEqualsIfExists",
        "ForAnyValueBinaryEquals",
        "ForAnyValueBinaryEqualsIfExists",
        each_item=True,
        pre=True,
    )
    def validate_binary(cls, value: Any) -> bytearray:
        try:
            value = b64decode(value)
        except binascii.Error:
            raise ValueError("Binary value not valid")
        return value

    def eval(self, values):
        if self._eval is None:
            self._eval = self.build_eval(self.dict())
        return self._eval(values)

    @classmethod
    def build_eval(cls, values: Dict) -> Dict:
        conditions_lambdas = [
            build_root_evaluator(function=key, arguments=value) for key, value in values.items() if value is not None
        ]
        return lambda kwargs: all(condition(kwargs) for condition in conditions_lambdas)

    def __call__(self, kwargs) -> Optional[bool]:
        try:
            return self.eval(kwargs)
        except Exception:
            logger.exception("Error raised while evaluating condition")
            return None

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, self.__class__):
            return self.dict(exclude_unset={"eval"}) == other.dict(exclude_unset={"eval"})
        else:
            return self.dict() == other



from pycfmodel.model.resources.properties.property import Property
from pycfmodel.model.types import ResolvableStr


class Tag(Property):
    """
    Tags for identifying and categorizing AWS resources. These are key-value pairs.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-resource-tags.html)
    """

    Key: ResolvableStr
    Value: ResolvableStr



import re

AWS_NOVALUE = "AWS::NoValue"

IMPLEMENTED_FUNCTIONS = {
    "Condition",
    "Fn::And",
    "Fn::Base64",
    "Fn::Equals",
    "Fn::FindInMap",
    "Fn::GetAtt",
    "Fn::GetAZs",
    "Fn::If",
    "Fn::ImportValue",
    "Fn::Join",
    "Fn::Not",
    "Fn::Or",
    "Fn::Select",
    "Fn::Split",
    "Fn::Sub",
    "Ref",
}


CONTAINS_STAR = re.compile(r"^.*[*].*$")
CONTAINS_CF_PARAM = re.compile(r"(\$\{[\w\:]+\})")
CONTAINS_SSM_PARAMETER = re.compile(r"{{resolve:ssm:([a-zA-Z0-9_./-]+:\d+)}}")

IPV4_ZERO_VALUE = "0.0.0.0/0"
IPV6_ZERO_VALUE = "::/0"



from typing import ClassVar, List, Optional

from pycfmodel.model.base import CustomModel
from pycfmodel.model.resources.properties.policy import Policy
from pycfmodel.model.resources.resource import Resource
from pycfmodel.model.types import Resolvable, ResolvableStr
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


class IAMGroupProperties(CustomModel):
    """
    Properties:

    - GroupName: Name of the group.
    - ManagedPolicyArns: ARN of the IAM policies to attach.
    - Path: Path to the group. See [IAM Identifiers](https://docs.aws.amazon.com/IAM/latest/UserGuide/Using_Identifiers.html).
    - Policies: Inline policies embedded in the IAM group.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-iam-group.html)
    """

    GroupName: Optional[ResolvableStr] = None
    ManagedPolicyArns: Optional[Resolvable[List[ResolvableStr]]] = None
    Path: Optional[ResolvableStr] = None
    Policies: Optional[Resolvable[List[Resolvable[Policy]]]] = None


class IAMGroup(Resource):
    """
    Properties:

    - Properties: A [IAM Group properties][pycfmodel.model.resources.iam_group.IAMGroupProperties] object.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-iam-group.html)
    """

    TYPE_VALUE: ClassVar = "AWS::IAM::Group"
    Type: str = TYPE_VALUE
    Properties: Optional[Resolvable[IAMGroupProperties]]

    @property
    def policy_documents(self) -> List[OptionallyNamedPolicyDocument]:
        result = []
        policies = self.Properties.Policies if self.Properties and self.Properties.Policies else []
        for policy in policies:
            result.append(OptionallyNamedPolicyDocument(name=policy.PolicyName, policy_document=policy.PolicyDocument))
        return result



from typing import ClassVar, List, Optional, Union

from pycfmodel.model.base import CustomModel
from pycfmodel.model.resources.properties.security_group_egress_prop import SecurityGroupEgressProp
from pycfmodel.model.resources.properties.security_group_ingress_prop import SecurityGroupIngressProp
from pycfmodel.model.resources.properties.tag import Tag
from pycfmodel.model.resources.resource import Resource
from pycfmodel.model.types import Resolvable, ResolvableStr


class SecurityGroupProperties(CustomModel):
    """
    Properties:

    - GroupDescription: Description for the security group.
    - GroupName: Name of the security group.
    - SecurityGroupEgress: Outbound rules associated with the security group.
    - SecurityGroupIngress: Inbound rules associated with the security group.
    - Tags: Array of key-value pairs.
    - VpcId: ID of the VPC for the security group.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ec2-security-group.html)
    """

    GroupDescription: ResolvableStr
    GroupName: Optional[ResolvableStr] = None
    SecurityGroupEgress: Optional[
        Resolvable[Union[SecurityGroupEgressProp, List[Resolvable[SecurityGroupEgressProp]]]]
    ] = None
    SecurityGroupIngress: Optional[
        Resolvable[Union[SecurityGroupIngressProp, List[Resolvable[SecurityGroupIngressProp]]]]
    ] = None
    Tags: Optional[Resolvable[List[Tag]]] = None
    VpcId: Optional[ResolvableStr] = None


class SecurityGroup(Resource):
    """
    Properties:

    - Properties: A [Security Group Properties][pycfmodel.model.resources.security_group.SecurityGroupProperties] object.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-ec2-security-group.html)
    """

    TYPE_VALUE: ClassVar = "AWS::EC2::SecurityGroup"
    Type: str = TYPE_VALUE
    Properties: Resolvable[SecurityGroupProperties]



import pytest

from pycfmodel.model.resources.iam_role import IAMRole
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


@pytest.fixture()
def iam_role():
    return IAMRole(
        **{
            "Type": "AWS::IAM::Role",
            "Properties": {
                "AssumeRolePolicyDocument": {
                    "Version": "2012-10-17",
                    "Statement": {
                        "Effect": "Allow",
                        "Principal": {"Service": ["ec2.amazonaws.com"], "AWS": "arn:aws:iam::111111111111:root"},
                        "Action": ["sts:AssumeRole"],
                    },
                },
                "Path": "/",
                "Policies": [
                    {
                        "PolicyName": "root",
                        "PolicyDocument": {
                            "Version": "2012-10-17",
                            "Statement": {"Effect": "Allow", "Action": "*", "Resource": "*"},
                        },
                    }
                ],
            },
        }
    )


def test_policies(iam_role):
    policies = iam_role.Properties.Policies
    assert len(policies) == 1
    assert policies[0].PolicyName == "root"


def test_iamrole_policy_documents(iam_role):
    assert iam_role.policy_documents == [
        OptionallyNamedPolicyDocument(name="root", policy_document=iam_role.Properties.Policies[0].PolicyDocument)
    ]



from typing import ClassVar, Optional

from pycfmodel.model.resources.properties.security_group_egress_prop import SecurityGroupEgressProp
from pycfmodel.model.resources.resource import Resource
from pycfmodel.model.types import ResolvableStr


class SecurityGroupEgressProperties(SecurityGroupEgressProp):
    """
    Properties:

    - CidrIp: IPv4 address range.
    - CidrIpv6: IPv6 address range.
    - Description: Description for the security group rule.
    - DestinationPrefixListId: The prefix list IDs for an AWS service.
    - DestinationSecurityGroupId: ID of the destination VPC security group.
    - FromPort: Start of port range for the TCP and UDP protocols, or an ICMP/ICMPv6 type number. A value of -1 indicates all ICMP/ICMPv6 types.
    - GroupId: ID of the security group.
    - IpProtocol: IP protocol name.
    - ToPort: End of port range for the TCP and UDP protocols, or an ICMP/ICMPv6 code. A value of -1 indicates all ICMP/ICMPv6 codes.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-security-group-egress.html)
    """

    GroupId: Optional[ResolvableStr] = None


class SecurityGroupEgress(Resource):
    """
    Properties:

    - Properties: A [Security Group Egress Properties][pycfmodel.model.resources.security_group_egress.SecurityGroupEgressProperties] object.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-security-group-egress.html)
    """

    TYPE_VALUE: ClassVar = "AWS::EC2::SecurityGroupEgress"
    Type: str = TYPE_VALUE
    Properties: SecurityGroupEgressProperties

    def ipv4_slash_zero(self) -> bool:
        return self.Properties.ipv4_slash_zero()

    def ipv6_slash_zero(self) -> bool:
        return self.Properties.ipv6_slash_zero()



import re
from ipaddress import IPv4Network, IPv6Network
from typing import Any, List, Pattern, Union

from pycfmodel.constants import IMPLEMENTED_FUNCTIONS


def is_resolvable_dict(value: Any) -> bool:
    return isinstance(value, dict) and len(value) == 1 and next(iter(value)) in IMPLEMENTED_FUNCTIONS


def convert_to_list(item: Union[Any, List[Any]]) -> List[Any]:
    if isinstance(item, list):
        return item
    return [item]


def regex_from_cf_string(action: str) -> Pattern:
    # Replace *
    action = action.replace("*", ".*")

    # Replace ?
    action = action.replace("?", ".{1}")

    return re.compile(f"^{action}$", re.IGNORECASE)


def not_ip(arg: Any) -> bool:
    return not isinstance(arg, IPv4Network) and not isinstance(arg, IPv6Network)



import logging
from typing import ClassVar

from pydantic import Extra, validator

from pycfmodel.model.resources.resource import Resource
from pycfmodel.model.resources.types import ResourceModels

logger = logging.getLogger(__file__)

_EXISTING_RESOURCE_TYPES = {klass.TYPE_VALUE for klass in ResourceModels.__args__}


class GenericResource(Resource):
    """This class is used for all resource types that we haven't had time to implement yet"""

    ALLOW_EXISTING_TYPES: ClassVar[bool] = True
    Type: str

    class Config(Resource.Config):
        extra = Extra.allow

    @validator("Type", pre=True)
    def check_type(cls, value, values, **kwargs):
        if value in _EXISTING_RESOURCE_TYPES:
            if cls.ALLOW_EXISTING_TYPES:
                logger.warning(f"Instantiating a GenericResource from a {value} in {values}")
            else:
                raise ValueError(f"Instantiation of GenericResource from {value} in {values} not allowed")
        return value



import pytest

from pycfmodel.model.resources.iam_user import IAMUser
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


@pytest.fixture()
def iam_user():
    return IAMUser(
        **{
            "Type": "AWS::IAM::User",
            "Properties": {
                "Path": "/",
                "UserName": "TestUser",
                "LoginProfile": {"Password": "ThisMyBestP@ssword", "PasswordResetRequired": False},
                "Policies": [
                    {
                        "PolicyName": "BadPolicy",
                        "PolicyDocument": {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Effect": "Allow",
                                    "Action": ["lambda:AddPermission", "ssm:SendCommand", "kms:Decrypt"],
                                    "Resource": "*",
                                }
                            ],
                        },
                    }
                ],
            },
        }
    )


def test_policies(iam_user):
    assert len(iam_user.Properties.Policies) == 1
    assert iam_user.Properties.Policies[0].PolicyName == "BadPolicy"


def test_iam_role_policy_documents(iam_user):
    assert iam_user.policy_documents == [
        OptionallyNamedPolicyDocument(name="BadPolicy", policy_document=iam_user.Properties.Policies[0].PolicyDocument)
    ]


def test_credential_check(iam_user):
    assert iam_user.has_hardcoded_credentials()



import pytest

from pycfmodel.model.resources.security_group_egress import SecurityGroupEgress


@pytest.fixture()
def security_group_egress_ipv4_1():
    return SecurityGroupEgress(
        **{
            "Type": "AWS::EC2::SecurityGroupEgress",
            "Properties": {
                "GroupId": "sg-12341234",
                "CidrIp": "1.1.1.1/0",
                "FromPort": 41,
                "ToPort": 45,
                "IpProtocol": "tcp",
            },
        }
    )


@pytest.fixture()
def security_group_egress_ipv4_2():
    return SecurityGroupEgress(
        **{
            "Type": "AWS::EC2::SecurityGroupEgress",
            "Properties": {
                "GroupId": "sg-2345",
                "CidrIp": "172.0.0.0",
                "FromPort": 41,
                "ToPort": 45,
                "IpProtocol": "tcp",
            },
        }
    )


@pytest.fixture()
def security_group_egress_ipv6():
    return SecurityGroupEgress(
        **{
            "Type": "AWS::EC2::SecurityGroupEgress",
            "Properties": {
                "GroupId": "sg-12341234",
                "CidrIpv6": "fc00::/7",
                "FromPort": 41,
                "ToPort": 45,
                "IpProtocol": "tcp",
            },
        }
    )


def test_security_group_egress(security_group_egress_ipv4_1):
    assert security_group_egress_ipv4_1.Properties.GroupId == "sg-12341234"
    assert isinstance(security_group_egress_ipv4_1.Properties.FromPort, int)


def test_slash_zero4(security_group_egress_ipv4_1):
    assert security_group_egress_ipv4_1.ipv4_slash_zero() is True


def test_not_slash_zero4(security_group_egress_ipv4_2):
    assert security_group_egress_ipv4_2.ipv4_slash_zero() is False


def test_not_slash_zero6(security_group_egress_ipv6):
    assert security_group_egress_ipv6.ipv6_slash_zero() is False



from typing import ClassVar, List, Optional

from pycfmodel.model.base import CustomModel
from pycfmodel.model.resources.properties.policy_document import PolicyDocument
from pycfmodel.model.resources.properties.tag import Tag
from pycfmodel.model.resources.resource import Resource
from pycfmodel.model.types import Resolvable, ResolvableBool, ResolvableInt, ResolvableStr
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


class KMSKeyProperties(CustomModel):
    """
    Properties:

    - Description: Description of the CMK.
    - EnableKeyRotation: Enables automatic rotation of the key for the customer master key.
    - Enabled: Specifies whether the customer master key (CMK) is enabled.
    - KeyPolicy: A [policy document][pycfmodel.model.resources.properties.policy_document.PolicyDocument] object.
    - KeyUsage: Determines the cryptographic operations.
    - KeySpec: Specifies the type of CMK to create.
    - MultiRegion: Allows multi-Region primary CMK to be replicated in other AWS Regions.
    - PendingWindowInDays: Number of days in the waiting period before AWS KMS deletes a CMK that has been removed from a CloudFormation stack.
    - Tags: Array of key-value pairs.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-kms-key.html)
    """

    Description: Optional[ResolvableStr] = None
    EnableKeyRotation: Optional[ResolvableBool] = None
    Enabled: Optional[ResolvableBool] = None
    KeyPolicy: Resolvable[PolicyDocument]
    KeyUsage: Optional[ResolvableStr] = None
    KeySpec: Optional[ResolvableStr] = None
    MultiRegion: Optional[ResolvableBool] = None
    PendingWindowInDays: Optional[ResolvableInt] = None
    Tags: Optional[Resolvable[List[Tag]]] = None


class KMSKey(Resource):
    """
    Properties:

    - Properties: A [KMS Key properties][pycfmodel.model.resources.kms_key.KMSKeyProperties] object.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-kms-key.html)
    """

    TYPE_VALUE: ClassVar = "AWS::KMS::Key"
    Type: str = TYPE_VALUE
    Properties: Resolvable[KMSKeyProperties]

    @property
    def policy_documents(self) -> List[OptionallyNamedPolicyDocument]:
        # Key policies are a bit different than most policies and would raise issues if treated equally
        return []



from pydantic import BaseModel, Extra, root_validator

from pycfmodel.utils import is_resolvable_dict


class CustomModel(BaseModel):
    class Config(BaseModel.Config):
        extra = Extra.forbid
        underscore_attrs_are_private = True

    def dict(self, *args, exclude_unset=True, **kwargs):
        return super().dict(*args, **kwargs, exclude_unset=exclude_unset)


class FunctionDict(BaseModel):
    # Inheriting directly from base model as we want to allow extra fields
    # and there are no default values that we need to suppress
    class Config(BaseModel.Config):
        extra = Extra.allow

    @root_validator(pre=True)
    def check_if_valid_function(cls, values):
        if not is_resolvable_dict(values):
            raise ValueError("FunctionDict should only have 1 key and be a function")
        return values



import pytest

from pycfmodel.model.resources.s3_bucket_policy import S3BucketPolicy
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


@pytest.fixture()
def s3_bucket_policy():
    return S3BucketPolicy(
        **{
            "Type": "AWS::S3::BucketPolicy",
            "Properties": {
                "Bucket": {"Ref": "S3Bucket"},
                "PolicyDocument": {
                    "Statement": [
                        {
                            "Action": ["*"],
                            "Effect": "Allow",
                            "Resource": "arn:aws:s3:::fakebucketfakebucket/*",
                            "Principal": {"AWS": ["156460612806"]},
                        }
                    ]
                },
            },
        }
    )


def test_policy_document(s3_bucket_policy):
    assert s3_bucket_policy.Properties.PolicyDocument.Statement[0].Action[0] == "*"


def test_s3_bucketpolicy_policy_document_property(s3_bucket_policy):
    assert s3_bucket_policy.policy_documents == [
        OptionallyNamedPolicyDocument(name=None, policy_document=s3_bucket_policy.Properties.PolicyDocument)
    ]



from typing import ClassVar, List

from pycfmodel.model.base import CustomModel
from pycfmodel.model.resources.properties.policy_document import PolicyDocument
from pycfmodel.model.resources.resource import Resource
from pycfmodel.model.types import Resolvable, ResolvableStr
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


class S3BucketPolicyProperties(CustomModel):
    """
    Properties:

    - Bucket: Name of the Amazon S3 bucket to which the policy applies.
    - PolicyDocument: A [policy document][pycfmodel.model.resources.properties.policy_document.PolicyDocument] object.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-s3-policy.html)
    """

    Bucket: ResolvableStr
    PolicyDocument: Resolvable[PolicyDocument]


class S3BucketPolicy(Resource):
    """
    Properties:

    - Properties: A [S3 Bucket Policy Properties][pycfmodel.model.resources.s3_bucket_policy.S3BucketPolicyProperties] object.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-s3-policy.html)
    """

    TYPE_VALUE: ClassVar = "AWS::S3::BucketPolicy"
    Type: str = TYPE_VALUE
    Properties: Resolvable[S3BucketPolicyProperties]

    @property
    def policy_documents(self) -> List[OptionallyNamedPolicyDocument]:
        return [OptionallyNamedPolicyDocument(name=None, policy_document=self.Properties.PolicyDocument)]



from dataclasses import dataclass
from typing import Optional

from pycfmodel.model.resources.properties.policy_document import PolicyDocument


@dataclass
class OptionallyNamedPolicyDocument:
    name: Optional[str]
    policy_document: PolicyDocument



import pytest

from pycfmodel.model.resources.iam_group import IAMGroup
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


@pytest.fixture()
def iam_group():
    return IAMGroup(
        **{
            "Type": "AWS::IAM::Group",
            "Properties": {
                "Path": "/",
                "GroupName": "GroupName",
                "Policies": [
                    {
                        "PolicyName": "BadPolicy",
                        "PolicyDocument": {
                            "Version": "2012-10-17",
                            "Statement": [
                                {
                                    "Effect": "Allow",
                                    "Action": ["lambda:AddPermission", "ssm:SendCommand", "kms:Decrypt"],
                                    "Resource": "*",
                                }
                            ],
                        },
                    }
                ],
            },
        }
    )


def test_policies(iam_group):
    policies = iam_group.Properties.Policies
    assert len(policies) == 1
    assert policies[0].PolicyName == "BadPolicy"


def test_iamgroup_policy_documents(iam_group):
    assert iam_group.policy_documents == [
        OptionallyNamedPolicyDocument(name="BadPolicy", policy_document=iam_group.Properties.Policies[0].PolicyDocument)
    ]



from typing import ClassVar, List

from pycfmodel.model.base import CustomModel
from pycfmodel.model.resources.properties.policy_document import PolicyDocument
from pycfmodel.model.resources.resource import Resource
from pycfmodel.model.types import Resolvable, ResolvableStr
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


class SQSQueuePolicyProperties(CustomModel):
    """
    Properties:

    - PolicyDocument: A [policy document][pycfmodel.model.resources.properties.policy_document.PolicyDocument] object.
    - Queues: URLs of the queues to add the policy.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-sqs-policy.html)
    """

    PolicyDocument: Resolvable[PolicyDocument]
    Queues: Resolvable[List[ResolvableStr]]


class SQSQueuePolicy(Resource):
    """
    Properties:

    - Properties: A [SQS Queue Policy Properties][pycfmodel.model.resources.sqs_queue_policy.SQSQueuePolicy] object.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-sqs-policy.html)
    """

    TYPE_VALUE: ClassVar = "AWS::SQS::QueuePolicy"
    Type: str = TYPE_VALUE
    Properties: Resolvable[SQSQueuePolicyProperties]

    @property
    def policy_documents(self) -> List[OptionallyNamedPolicyDocument]:
        return [OptionallyNamedPolicyDocument(name=None, policy_document=self.Properties.PolicyDocument)]



from typing import ClassVar, List, Optional

from pycfmodel.model.base import CustomModel
from pycfmodel.model.resources.properties.policy_document import PolicyDocument
from pycfmodel.model.resources.resource import Resource
from pycfmodel.model.types import Resolvable, ResolvableStr, ResolvableStrOrList
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


class IAMPolicyProperties(CustomModel):
    """
    Properties:

    - Groups: Friendly name of the IAM groups to attach the policy to.
    - PolicyDocument: A [policy document][pycfmodel.model.resources.properties.policy_document.PolicyDocument] object.
    - PolicyName: Name of the policy.
    - Roles: Friendly name of the IAM roles to attach the policy to.
    - Users: Friendly name of the IAM users to attach the policy to.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-iam-policy.html)
    """

    Groups: Optional[ResolvableStrOrList] = None
    PolicyDocument: Resolvable[PolicyDocument]
    PolicyName: ResolvableStr
    Roles: Optional[ResolvableStrOrList] = None
    Users: Optional[ResolvableStrOrList] = None


class IAMPolicy(Resource):
    """
    Properties:

    - Properties: A [IAM Policy properties][pycfmodel.model.resources.iam_policy.IAMPolicyProperties] object.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-iam-policy.html)
    """

    TYPE_VALUE: ClassVar = "AWS::IAM::Policy"
    Type: str = TYPE_VALUE
    Properties: Resolvable[IAMPolicyProperties]

    @property
    def policy_documents(self) -> List[OptionallyNamedPolicyDocument]:
        return [
            OptionallyNamedPolicyDocument(
                name=self.Properties.PolicyName, policy_document=self.Properties.PolicyDocument
            )
        ]



import pytest

from pycfmodel.model.resources.ec2_vpc_endpoint_policy import EC2VpcEndpointPolicy
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


@pytest.fixture()
def ec2_vpc_endpoint_policy():
    return EC2VpcEndpointPolicy(
        **{
            "Type": "AWS::EC2::VPCEndpoint",
            "Properties": {
                "PolicyDocument": {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Principal": "*",
                            "Action": ["s3:GetObject"],
                            "Resource": "*",
                        }
                    ],
                },
                "RouteTableIds": [{"Ref": "routetableA"}, {"Ref": "routetableB"}],
                "ServiceName": "com.amazonaws.eu-west-1.s3",
                "VpcId": "vpc-123-abc",
            },
        }
    )


@pytest.fixture()
def ec2_vpc_endpoint_policy_no_policy_document():
    return EC2VpcEndpointPolicy(
        **{
            "Type": "AWS::EC2::VPCEndpoint",
            "Properties": {
                "SecurityGroupIds": [{"Ref": "VPCEndpointSecurityGroup"}],
                "ServiceName": {
                    "Fn::Sub": [
                        "com.amazonaws.vpce.${AWS::Region}.${ServiceID}",
                        {"ServiceID": {"Ref": "VPCEndpointService"}},
                    ]
                },
                "SubnetIds": ["subnet-123", "subnet-456"],
                "VpcEndpointType": "Interface",
                "VpcId": "vpc-abc987",
            },
        }
    )


def test_ec2_vpc_endpoint(ec2_vpc_endpoint_policy: EC2VpcEndpointPolicy):
    assert ec2_vpc_endpoint_policy.Properties.ServiceName == "com.amazonaws.eu-west-1.s3"
    assert ec2_vpc_endpoint_policy.Properties.VpcId == "vpc-123-abc"
    assert len(ec2_vpc_endpoint_policy.Properties.RouteTableIds) == 2
    assert ec2_vpc_endpoint_policy.Properties.PolicyDocument.Statement[0].Effect == "Allow"
    assert ec2_vpc_endpoint_policy.Properties.PolicyDocument.Statement[0].Resource == "*"


def test_ec2_vpc_endpoint_policy_documents(ec2_vpc_endpoint_policy):
    assert ec2_vpc_endpoint_policy.policy_documents == [
        OptionallyNamedPolicyDocument(name=None, policy_document=ec2_vpc_endpoint_policy.Properties.PolicyDocument)
    ]


def test_ec2_vpc_endpoint_policy_documents_no_document(ec2_vpc_endpoint_policy_no_policy_document):
    assert ec2_vpc_endpoint_policy_no_policy_document.policy_documents == []



from typing import List, Union

from pycfmodel.cloudformation_actions import CLOUDFORMATION_ACTIONS
from pycfmodel.utils import regex_from_cf_string


def _expand_action(action: str, not_action=False) -> List[str]:
    if isinstance(action, str):
        pattern = regex_from_cf_string(action)
        if not_action:
            return sorted(
                set(CLOUDFORMATION_ACTIONS) - set(action for action in CLOUDFORMATION_ACTIONS if pattern.match(action))
            )
        return sorted(set(action for action in CLOUDFORMATION_ACTIONS if pattern.match(action)))

    raise ValueError(f"Not supported type: {type(action)}")


def _expand_actions(actions: Union[str, List[str]], not_action=False) -> List[str]:
    if isinstance(actions, str):
        return _expand_action(actions, not_action=not_action)

    if isinstance(actions, list):
        expanded_actions = set()
        for action in actions:
            expanded_actions.update(_expand_action(action))
        if not_action:
            return sorted(set(CLOUDFORMATION_ACTIONS) - expanded_actions)
        return sorted(expanded_actions)

    raise ValueError(f"Not supported type: {type(actions)}")


def expand_actions(obj):
    if isinstance(obj, dict):
        for key, value in obj.items():
            if key == "Action":
                obj[key] = _expand_actions(value)
            elif key == "NotAction":
                obj[key] = _expand_actions(value, not_action=True)
            else:
                obj[key] = expand_actions(value)

    if isinstance(obj, list):
        obj = [expand_actions(value) for value in obj]

    return obj



import pytest
from pydantic import ValidationError

from pycfmodel.model.resources.s3_bucket import S3Bucket


@pytest.fixture()
def valid_s3_bucket():
    return S3Bucket(
        **{
            "Type": "AWS::S3::Bucket",
            "Properties": {
                "AccessControl": "Private",
                "VersioningConfiguration": {"Status": "Enabled"},
                "Tags": [{"Key": "Project", "Value": "test"}],
                "ReplicationConfiguration": {
                    "Role": "arn:aws:iam::123456789012:role/replication_role",
                    "Rules": [
                        {
                            "Id": "MyRule1",
                            "Status": "Enabled",
                            "Prefix": "MyPrefix",
                            "Destination": {
                                "Bucket": "arn:aws:s3:::my-replication-bucket",
                                "StorageClass": "STANDARD",
                            },
                        },
                        {
                            "Status": "Enabled",
                            "Prefix": "MyOtherPrefix",
                            "Destination": {"Bucket": "arn:aws:s3:::my-replication-bucket"},
                        },
                    ],
                },
            },
        }
    )


def test_valid_s3_bucket_resource(valid_s3_bucket):
    assert (
        valid_s3_bucket.Properties.ReplicationConfiguration.get("Role")
        == "arn:aws:iam::123456789012:role/replication_role"
    )
    assert valid_s3_bucket.Properties.ReplicationConfiguration.get("Rules")[1].get("Status") == "Enabled"
    assert valid_s3_bucket.Properties.VersioningConfiguration == {"Status": "Enabled"}
    assert valid_s3_bucket.Properties.Tags[0].Value == "test"


def test_extra_fields_not_allowed_s3_bucket():
    with pytest.raises(ValidationError) as exc_info:
        S3Bucket(
            **{
                "Type": "AWS::S3::Bucket",
                "Properties": {"AccelerateConfiguration": "None", "AnalyticsConfigurations": {"a": "b"}, "foo": "bar"},
            }
        )

    assert exc_info.value.errors() == [
        {
            "loc": ("Properties", "AccelerateConfiguration"),
            "msg": "value is not a valid dict",
            "type": "type_error.dict",
        },
        {
            "loc": ("Properties", "AccelerateConfiguration"),
            "msg": "value is not a valid dict",
            "type": "type_error.dict",
        },
        {
            "loc": ("Properties", "AnalyticsConfigurations"),
            "msg": "value is not a valid list",
            "type": "type_error.list",
        },
        {
            "loc": ("Properties", "AnalyticsConfigurations", "__root__"),
            "msg": "FunctionDict should only have 1 key and be a function",
            "type": "value_error",
        },
        {"loc": ("Properties", "foo"), "msg": "extra fields not permitted", "type": "value_error.extra"},
        {
            "loc": ("Properties", "__root__"),
            "msg": "FunctionDict should only have 1 key and be a function",
            "type": "value_error",
        },
    ]



import pytest

from pycfmodel.model.resources.sqs_queue_policy import SQSQueuePolicy
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


@pytest.fixture()
def sqs_queue_policy():
    return SQSQueuePolicy(
        **{
            "Type": "AWS::SQS::QueuePolicy",
            "Properties": {
                "Queues": [{"Ref": "queue1"}],
                "PolicyDocument": {
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "NotAction": ["sqs:Break*"],
                            "Principal": {"AWS": "arn:aws:iam::111111111111:user/dave.mustaine"},
                            "Resource": "*",
                        }
                    ]
                },
            },
        }
    )


def test_sqs_queue(sqs_queue_policy):
    assert len(sqs_queue_policy.Properties.Queues) == 1
    assert sqs_queue_policy.Properties.PolicyDocument.Statement[0].Effect == "Allow"
    assert sqs_queue_policy.Properties.PolicyDocument.Statement[0].NotAction[0] == "sqs:Break*"


def test_sqs_policy_documents(sqs_queue_policy):
    assert sqs_queue_policy.policy_documents == [
        OptionallyNamedPolicyDocument(name=None, policy_document=sqs_queue_policy.Properties.PolicyDocument)
    ]



from typing import ClassVar, List, Optional

from pycfmodel.model.base import CustomModel
from pycfmodel.model.resources.properties.policy_document import PolicyDocument as resource_policy_document
from pycfmodel.model.resources.resource import Resource
from pycfmodel.model.types import Resolvable, ResolvableBool, ResolvableStr
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


class EC2VpcEndpointPolicyProperties(CustomModel):
    """
    Properties:

    - PolicyDocument: A [policy document][pycfmodel.model.resources.properties.policy_document.PolicyDocument] object.
    - PrivateDnsEnabled: Indicate whether to associate a private hosted zone with the specified VPC.
    - RouteTableIds: One or more route table IDs.
    - SecurityGroupIds: The ID of one or more security groups to associate with the endpoint network interface.
    - ServiceName: The service name.
    - SubnetIds: The ID of one or more subnets in which to create an endpoint network interface.
    - VpcEndpointType: The type of endpoint.
    - VpcId: The ID of the VPC in which the endpoint will be used.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-vpcendpoint.html)
    """

    PolicyDocument: Optional[Resolvable[resource_policy_document]] = None
    PrivateDnsEnabled: Optional[ResolvableBool] = None
    RouteTableIds: Optional[Resolvable[List[ResolvableStr]]] = None
    SecurityGroupIds: Optional[Resolvable[List[ResolvableStr]]] = None
    ServiceName: ResolvableStr
    SubnetIds: Optional[Resolvable[List[ResolvableStr]]] = None
    VpcEndpointType: Optional[ResolvableStr] = None
    VpcId: ResolvableStr


class EC2VpcEndpointPolicy(Resource):
    """
    Properties:

    - Properties: An [EC2 VPC Policy Endpoint Properties][pycfmodel.model.resources.ec2_vpc_endpoint_policy.EC2VpcEndpointPolicyProperties]
    object.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-vpcendpoint.html)
    """

    TYPE_VALUE: ClassVar = "AWS::EC2::VPCEndpoint"
    Type: str = TYPE_VALUE
    Properties: Resolvable[EC2VpcEndpointPolicyProperties]

    @property
    def policy_documents(self) -> List[OptionallyNamedPolicyDocument]:
        if not self.Properties.PolicyDocument:
            return []
        return [OptionallyNamedPolicyDocument(name=None, policy_document=self.Properties.PolicyDocument)]



from typing import ClassVar, List, Optional

from pycfmodel.model.base import CustomModel
from pycfmodel.model.resources.properties.policy import Policy
from pycfmodel.model.resources.properties.policy_document import PolicyDocument
from pycfmodel.model.resources.resource import OptionallyNamedPolicyDocument, Resource
from pycfmodel.model.types import Resolvable, ResolvableIntOrStr, ResolvableStr


class IAMRoleProperties(CustomModel):
    """
    Properties:

    - AssumeRolePolicyDocument: A [policy document][pycfmodel.model.resources.properties.policy_document.PolicyDocument] object.
    - ManagedPolicyArns: List of ARNs of the IAM managed policies to attach.
    - MaxSessionDuration: Maximum session duration (in seconds).
    - Path: Path to the role.
    - PermissionsBoundary: ARN of the policy used to set the permissions boundary.
    - Policies: A list of [policy][pycfmodel.model.resources.properties.policy.Policy] objects.
    - RoleName: Name of the role.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-iam-role.html)
    """

    AssumeRolePolicyDocument: Resolvable[PolicyDocument]
    ManagedPolicyArns: Optional[Resolvable[List[ResolvableStr]]] = None
    MaxSessionDuration: Optional[ResolvableIntOrStr] = None
    Path: Optional[ResolvableStr] = None
    PermissionsBoundary: Optional[ResolvableStr] = None
    Policies: Optional[Resolvable[List[Resolvable[Policy]]]] = None
    RoleName: Optional[ResolvableStr] = None


class IAMRole(Resource):
    """
    Properties:

    - Properties: A [IAM Role properties][pycfmodel.model.resources.iam_role.IAMRoleProperties] object.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-iam-role.html)
    """

    TYPE_VALUE: ClassVar = "AWS::IAM::Role"
    Type: str = TYPE_VALUE
    Properties: Resolvable[IAMRoleProperties]

    @property
    def policy_documents(self) -> List[OptionallyNamedPolicyDocument]:
        result = []
        policies = self.Properties.Policies if self.Properties and self.Properties.Policies else []
        for policy in policies:
            result.append(OptionallyNamedPolicyDocument(name=policy.PolicyName, policy_document=policy.PolicyDocument))
        return result



from typing import ClassVar, Dict, List, Optional

from pydantic import validator

from pycfmodel.model.base import CustomModel
from pycfmodel.model.parameter import Parameter
from pycfmodel.model.types import ResolvableCondition, ResolvableStr, ResolvableStrOrList
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


class Resource(CustomModel):
    TYPE_VALUE: ClassVar[str]
    Type: str
    Condition: Optional[ResolvableCondition] = None
    CreatePolicy: Optional[Dict] = None
    DeletionPolicy: Optional[ResolvableStr] = None
    DependsOn: Optional[ResolvableStrOrList] = None
    Metadata: Optional[Dict] = None
    UpdatePolicy: Optional[Dict] = None
    UpdateReplacePolicy: Optional[ResolvableStr] = None

    @validator("Type")
    def check_type(cls, value):
        if value != cls.TYPE_VALUE:
            raise ValueError(f"Value needs to be {cls.TYPE_VALUE}")
        return value

    def has_hardcoded_credentials(self) -> bool:
        if not self.Metadata or not self.Metadata.get("AWS::CloudFormation::Authentication"):
            return False

        for auth in self.Metadata["AWS::CloudFormation::Authentication"].values():
            if not all(
                [
                    auth.get("accessKeyId", Parameter.NO_ECHO_NO_DEFAULT) == Parameter.NO_ECHO_NO_DEFAULT,
                    auth.get("password", Parameter.NO_ECHO_NO_DEFAULT) == Parameter.NO_ECHO_NO_DEFAULT,
                    auth.get("secretKey", Parameter.NO_ECHO_NO_DEFAULT) == Parameter.NO_ECHO_NO_DEFAULT,
                ]
            ):
                return True

        return False

    @property
    def policy_documents(self) -> List[OptionallyNamedPolicyDocument]:
        return []



from datetime import date
from typing import Any, ClassVar, Collection, Dict, List, Optional, Type, Union

from pycfmodel.action_expander import expand_actions
from pycfmodel.constants import AWS_NOVALUE
from pycfmodel.model.base import CustomModel
from pycfmodel.model.parameter import Parameter
from pycfmodel.model.resources.generic_resource import GenericResource
from pycfmodel.model.resources.resource import Resource
from pycfmodel.model.resources.types import ResourceModels
from pycfmodel.model.types import Resolvable
from pycfmodel.resolver import _extended_bool, resolve


class CFModel(CustomModel):
    """
    Template that describes AWS infrastructure.

    Properties:

    - AWSTemplateFormatVersion
    - Conditions: Conditions that control behaviour of the template.
    - Description: Description for the template.
    - Mappings: A 3 level mapping of keys and associated values.
    - Metadata: Additional information about the template.
    - Outputs: Output values of the template.
    - Parameters: Parameters to the template.
    - Resources: Stack resources and their properties.
    - Rules
    - Transform: For serverless applications, specifies the version of the AWS Serverless Application Model (AWS SAM) to use.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/template-anatomy.html)
    """

    AWSTemplateFormatVersion: Optional[date]
    Conditions: Optional[Dict] = {}
    Description: Optional[str] = None
    Mappings: Optional[Dict[str, Dict[str, Dict[str, Any]]]] = {}
    Metadata: Optional[Dict[str, Any]] = None
    Outputs: Optional[Dict[str, Dict[str, Union[str, Dict]]]] = {}
    Parameters: Optional[Dict[str, Parameter]] = {}
    Resources: Dict[str, Resolvable[Union[ResourceModels, GenericResource]]] = {}
    Rules: Optional[Dict] = {}
    Transform: Optional[Union[str, List[str]]]

    PSEUDO_PARAMETERS: ClassVar[Dict[str, Union[str, List[str]]]] = {
        # default pseudo parameters
        "AWS::AccountId": "123456789012",
        "AWS::NotificationARNs": [],
        "AWS::NoValue": AWS_NOVALUE,
        "AWS::Partition": "aws",
        "AWS::Region": "eu-west-1",
        "AWS::StackId": "",
        "AWS::StackName": "",
        "AWS::URLSuffix": "amazonaws.com",
    }

    def resolve(self, extra_params=None) -> "CFModel":
        """
        Resolve all intrinsic functions on the template.

        Arguments:
            extra_params: Values of parameters passed to the Cloudformation.

        Returns:
            A new CFModel.
        """
        extra_params = {} if extra_params is None else extra_params
        # default parameters
        params = {}
        for key, parameter in self.Parameters.items():
            passed_value = extra_params.pop(key, None)
            ref_value = parameter.get_ref_value(passed_value)
            if ref_value is not None:
                params[key] = ref_value

        extended_parameters = {**self.PSEUDO_PARAMETERS, **params, **extra_params}
        dict_value = self.dict()

        conditions = dict_value.pop("Conditions", {})
        resolved_conditions = {
            key: _extended_bool(resolve(value, extended_parameters, self.Mappings, {}))
            for key, value in conditions.items()
        }

        resources = dict_value.pop("Resources")
        resolved_resources = {
            key: resolve(value, extended_parameters, self.Mappings, resolved_conditions)
            for key, value in resources.items()
            if value.get("Condition") is None
            or (value.get("Condition") is not None and resolved_conditions.get(value["Condition"], True))
        }
        return CFModel(**dict_value, Conditions=resolved_conditions, Resources=resolved_resources)

    def expand_actions(self) -> "CFModel":
        """
        Returns a model which has expanded all wildcards (`*`) to get all implied actions for every resource.
        For example:\n
          - a model containing `s3:*` will be expanded to list all the possible S3 actions.
          - a model containing `s3:Get*` will be expanded to all the `Get*` actions only.

        This method can handle the cases of both `Action` and `NotAction`.

        [Known AWS Actions](https://github.com/Skyscanner/pycfmodel/blob/master/pycfmodel/cloudformation_actions.py).
        These known actions can be updated by executing:

        ```
        python3 scripts/generate_cloudformation_actions_file.py
        ```
        """
        dict_value = self.dict()

        resources = dict_value.pop("Resources")
        expanded_resources = {key: expand_actions(value) for key, value in resources.items()}

        return CFModel(**dict_value, Resources=expanded_resources)

    def resources_filtered_by_type(
        self, allowed_types: Collection[Union[str, Type[Resource]]]
    ) -> Dict[str, Dict[str, Resource]]:
        """
        Filtered resources based on types.

        Arguments:
            allowed_types: Collection of desired types.

        Returns:
            Dictionary where key is the logical id and value is the resource.
        """
        result = {}
        allowed_resource_classes = tuple(x for x in allowed_types if isinstance(x, type))
        for resource_name, resource in self.Resources.items():
            if isinstance(resource, allowed_resource_classes) or resource.Type in allowed_types:
                result[resource_name] = resource
        return result



import pytest

from pycfmodel.model.resources.iam_policy import IAMPolicy
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


@pytest.fixture()
def iam_policy():
    return IAMPolicy(
        **{
            "Type": "AWS::IAM::Policy",
            "Properties": {
                "PolicyName": "Policy",
                "PolicyDocument": {
                    "Version": "2012-10-17",
                    "Statement": [
                        {
                            "Effect": "Allow",
                            "Action": ["lambda:AddPermission", "ssm:SendCommand", "kms:Decrypt"],
                            "Resource": "*",
                        }
                    ],
                },
            },
        }
    )


def test_policies(iam_policy):
    assert "Policy" == iam_policy.Properties.PolicyName


def test_iam_managedpolicy_policy_documents(iam_policy):
    assert iam_policy.policy_documents == [
        OptionallyNamedPolicyDocument(name="Policy", policy_document=iam_policy.Properties.PolicyDocument)
    ]



import logging
from typing import Dict, List, Optional, Pattern, Union

from pydantic import validator

from pycfmodel.action_expander import _expand_action
from pycfmodel.model.resources.properties.property import Property
from pycfmodel.model.resources.properties.statement_condition import StatementCondition
from pycfmodel.model.types import ResolvableStr, ResolvableStrOrList
from pycfmodel.utils import is_resolvable_dict

logger = logging.getLogger(__file__)

PrincipalTypes = Union[ResolvableStr, List[ResolvableStr], Dict[str, Union[ResolvableStr, List[ResolvableStr]]]]


class Statement(Property):
    """
    Contains information about an attached policy.

    Properties:

    - Sid: Optional identifier.
    - Effect: Whether the statement results in an allow or an explicit deny.
    - Principal: Specify the IAM user, federated user, IAM role, AWS account, AWS service, or other principal that is allowed to access a resource.
    - NotPrincipal: Specify the IAM user, federated user, IAM role, AWS account, AWS service, or other principal that is not allowed or denied access to a resource.
    - Action: Specific action or actions that will be allowed or denied.
    - NotAction: Explicitly matches everything except the specified action or list of actions.
    - Resource: Specifies the object or objects that the statement covers.
    - NotResource: Specifies the object or objects that the statement does not cover.
    - Condition: Element to match the condition key and value in the policy against values in the request context.

    More info at [AWS Docs](https://docs.aws.amazon.com/IAM/latest/UserGuide/reference_policies_elements.html)
    """

    Sid: Optional[ResolvableStr] = None
    Effect: Optional[ResolvableStr] = None
    Principal: Optional[PrincipalTypes] = None
    NotPrincipal: Optional[PrincipalTypes] = None
    Action: Optional[ResolvableStrOrList] = None
    NotAction: Optional[ResolvableStrOrList] = None
    Resource: Optional[ResolvableStrOrList] = None
    NotResource: Optional[ResolvableStrOrList] = None
    Condition: Optional[StatementCondition] = None

    @validator("Effect")
    def capitalize_if_str(cls, v: ResolvableStr):
        if isinstance(v, str):
            return v.capitalize()
        return v

    def get_action_list(self, include_action=True, include_not_action=True) -> List[ResolvableStr]:
        """
        Gets all actions specified in `Action` and `NotAction`.

        Returns:
            List of actions.
        """
        action_list = []
        included_actions = []
        if include_action:
            included_actions.append(self.Action)
        if include_not_action:
            included_actions.append(self.NotAction)
        for actions in included_actions:
            if isinstance(actions, List):
                action_list.extend(actions)
            elif isinstance(actions, (str, dict)):
                action_list.append(actions)
        return action_list

    def get_expanded_action_list(self) -> List[str]:
        action_list = set()
        for action in self.get_action_list(include_action=True, include_not_action=False):
            action_list.update(_expand_action(action))

        for not_action in self.get_action_list(include_not_action=True, include_action=False):
            action_list.update(_expand_action(not_action, not_action=True))

        return sorted(action_list)

    def get_resource_list(self) -> List[ResolvableStr]:
        """
        Gets all resources specified in `Resource` and `NotResource`.

        Returns:
            List of resources.
        """
        resource_list = []
        for resources in [self.Resource, self.NotResource]:
            if isinstance(resources, List):
                resource_list.extend(resources)
            elif isinstance(resources, (str, dict)):
                resource_list.append(resources)
        return resource_list

    def get_principal_list(self) -> List[ResolvableStr]:
        """
        Gets all actions specified in `Principal` and `NotPrincipal`.

        Returns:
            List of principals.
        """
        principal_list = []
        for principals in [self.Principal, self.NotPrincipal]:
            if isinstance(principals, list):
                principal_list.extend(principals)
            elif isinstance(principals, str):
                principal_list.append(principals)
            elif is_resolvable_dict(principals):
                principal_list.append(principals)
            elif isinstance(principals, dict):
                for value in principals.values():
                    if isinstance(value, (str, Dict)):
                        principal_list.append(value)
                    elif isinstance(value, List):
                        principal_list.extend(value)
            elif principals is not None:
                raise ValueError(f"Not supported type: {type(principals)}")
        return principal_list

    def actions_with(self, pattern: Pattern) -> List[str]:
        """
        Finds all actions which match the pattern.

        Arguments:
            pattern: Pattern to match.

        Returns:
            List of actions.
        """
        return [action for action in self.get_action_list() if isinstance(action, str) and pattern.match(action)]

    def principals_with(self, pattern: Pattern) -> List[str]:
        """
        Finds all principals which match the pattern.

        Arguments:
            pattern: Pattern to match.

        Returns:
            List of principals.
        """
        return [
            principal
            for principal in self.get_principal_list()
            if isinstance(principal, str) and pattern.match(principal)
        ]

    def resources_with(self, pattern: Pattern) -> List[str]:
        """
        Finds all resources which match the pattern.

        Arguments:
            pattern: Pattern to match.

        Returns:
            List of resources.
        """
        return [
            resource for resource in self.get_resource_list() if isinstance(resource, str) and pattern.match(resource)
        ]

    def non_whitelisted_principals(self, whitelist: List[str]) -> List[str]:
        """
        Find non whitelisted principals.

        Arguments:
            whitelist: List of whitelisted principals.

        Returns:
            List of principals.
        """
        return [
            principal
            for principal in self.get_principal_list()
            if isinstance(principal, str) and principal not in whitelist
        ]



from typing import ClassVar, List, Optional

from pycfmodel.model.base import CustomModel
from pycfmodel.model.resources.properties.tag import Tag
from pycfmodel.model.resources.resource import Resource
from pycfmodel.model.types import Resolvable, ResolvableBool, ResolvableDict, ResolvableStr


class S3BucketProperties(CustomModel):
    """
    Properties:

    - AccelerateConfiguration: Configures the transfer acceleration state for an Amazon S3 bucket.
    - AccessControl: A canned access control list (ACL) that grants predefined permissions to the bucket.
    - AnalyticsConfigurations: Specifies the configuration and any analyses for the analytics filter of an Amazon S3 bucket.
    - BucketEncryption: Specifies encryption on the bucket.
    - BucketName: The name of the bucket.
    - CorsConfiguration: Describes the cross-origin access configuration for objects in an Amazon S3 bucket.
    - IntelligentTieringConfigurations: Defines how Amazon S3 handles Intelligent-Tiering storage.
    - InventoryConfigurations: Specifies the inventory configuration for an Amazon S3 bucket.
    - LifecycleConfiguration: Specifies the lifecycle configuration for objects in an Amazon S3 bucket.
    - LoggingConfiguration: Settings that define where logs are stored.
    - MetricsConfigurations: Specifies a metrics configuration for the CloudWatch request metrics.
    - NotificationConfiguration: Defines how Amazon S3 handles bucket notifications.
    - ObjectLockConfiguration: Places an Object Lock configuration on the specified bucket.
    - ObjectLockEnabled: Indicates whether this bucket has an Object Lock configuration enabled.
    - OwnershipControls: Defines how Amazon S3 handles object ownership rules.
    - PublicAccessBlockConfiguration: Defines how Amazon S3 handles public access.
    - ReplicationConfiguration: Configuration for replicating objects in an S3 bucket.
    - Tags: An arbitrary set of tags (key-value pairs) for the bucket.
    - VersioningConfiguration: If enabled, allows for multiple versions of all objects in this bucket to be stored.
    - WebsiteConfiguration: Information used to configure the bucket as a static website.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-s3-bucket.html)
    """

    AccelerateConfiguration: Optional[ResolvableDict] = None
    AccessControl: Optional[ResolvableStr] = None
    AnalyticsConfigurations: Optional[Resolvable[List[ResolvableDict]]] = None
    BucketEncryption: Optional[ResolvableDict] = None
    BucketName: Optional[ResolvableStr] = None
    CorsConfiguration: Optional[ResolvableDict] = None
    IntelligentTieringConfigurations: Optional[Resolvable[List[ResolvableDict]]] = None
    InventoryConfigurations: Optional[Resolvable[List[ResolvableDict]]] = None
    LifecycleConfiguration: Optional[ResolvableDict] = None
    LoggingConfiguration: Optional[ResolvableDict] = None
    MetricsConfigurations: Optional[Resolvable[List[ResolvableDict]]] = None
    NotificationConfiguration: Optional[ResolvableDict] = None
    ObjectLockConfiguration: Optional[ResolvableDict] = None
    ObjectLockEnabled: Optional[ResolvableBool] = None
    OwnershipControls: Optional[ResolvableDict] = None
    PublicAccessBlockConfiguration: Optional[ResolvableDict] = None
    ReplicationConfiguration: Optional[ResolvableDict] = None
    Tags: Optional[Resolvable[List[Tag]]] = None
    VersioningConfiguration: Optional[ResolvableDict] = None
    WebsiteConfiguration: Optional[ResolvableDict] = None


class S3Bucket(Resource):
    """
    Properties:

    - Properties: A [S3 Bucket Properties][pycfmodel.model.resources.s3_bucket.S3BucketProperties] object.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-s3-bucket.html)
    """

    TYPE_VALUE: ClassVar = "AWS::S3::Bucket"
    Type: str = TYPE_VALUE
    Properties: Resolvable[S3BucketProperties]



from typing import ClassVar, Dict, List, Optional

from pycfmodel.model.base import CustomModel
from pycfmodel.model.parameter import Parameter
from pycfmodel.model.resources.properties.policy import Policy
from pycfmodel.model.resources.resource import Resource
from pycfmodel.model.types import Resolvable, ResolvableStr
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


class IAMUserProperties(CustomModel):
    """
    Properties:

    - Groups: List of groups to attach.
    - LoginProfile: Name and password for the user.
    - ManagedPolicyArns: List of ARNs of the IAM managed policies to attach.
    - Path: Path to the user.
    - PermissionsBoundary: ARN of the policy used to set the permissions boundary.
    - Policies: A list of [policy][pycfmodel.model.resources.properties.policy.Policy] objects.
    - UserName: Name of the user.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-iam-user.html)
    """

    Groups: Optional[Resolvable[List[ResolvableStr]]] = None
    LoginProfile: Optional[Dict] = None
    ManagedPolicyArns: Optional[Resolvable[List[ResolvableStr]]] = None
    Path: Optional[ResolvableStr] = None
    PermissionsBoundary: Optional[ResolvableStr] = None
    Policies: Optional[Resolvable[List[Resolvable[Policy]]]] = None
    UserName: Optional[ResolvableStr] = None


class IAMUser(Resource):
    """
    Properties:

    - Properties: A [IAM User properties][pycfmodel.model.resources.iam_user.IAMUserProperties] object.

    More info at [AWS Docs](https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-iam-user.html)
    """

    TYPE_VALUE: ClassVar = "AWS::IAM::User"
    Type: str = TYPE_VALUE
    Properties: Optional[Resolvable[IAMUserProperties]]

    def has_hardcoded_credentials(self) -> bool:
        """Returns True if login profile password contains a hardcoded string, otherwise False."""
        if self.Properties:
            login_profile = self.Properties.LoginProfile
            if login_profile and login_profile.get("Password"):
                if login_profile["Password"] != Parameter.NO_ECHO_NO_DEFAULT:
                    return True

        return super().has_hardcoded_credentials()

    @property
    def policy_documents(self) -> List[OptionallyNamedPolicyDocument]:
        result = []
        policies = self.Properties.Policies if self.Properties and self.Properties.Policies else []
        for policy in policies:
            result.append(OptionallyNamedPolicyDocument(policy.PolicyName, policy.PolicyDocument))
        return result



import pytest

from pycfmodel.model.resources.security_group import SecurityGroup


@pytest.fixture()
def security_group():
    return SecurityGroup(
        **{
            "Type": "AWS::EC2::SecurityGroup",
            "Properties": {
                "GroupDescription": "some_group_desc",
                "SecurityGroupIngress": [{"CidrIp": "10.1.2.3/32", "FromPort": 34, "ToPort": 36, "IpProtocol": "tcp"}],
                "SecurityGroupEgress": [{"CidrIp": "10.1.2.3/32", "FromPort": 34, "ToPort": 36, "IpProtocol": "tcp"}],
                "Tags": [{"Key": "a", "Value": "b"}],
                "VpcId": "vpc-9f8e9dfa",
            },
        }
    )


def test_security_group(security_group):
    assert security_group.Properties.GroupDescription == "some_group_desc"
    assert security_group.Properties.Tags[0].Value == "b"
    assert not security_group.Properties.SecurityGroupIngress[0].ipv4_slash_zero()
    assert not security_group.Properties.SecurityGroupEgress[0].ipv4_slash_zero()



from typing import List, Optional, Pattern, Union

from pydantic import Extra

from pycfmodel.cloudformation_actions import CLOUDFORMATION_ACTIONS
from pycfmodel.model.resources.properties.property import Property
from pycfmodel.model.resources.properties.statement import Statement
from pycfmodel.model.types import Resolvable, ResolvableDate


class PolicyDocument(Property):
    """
    Contains information about an attached policy.

    Properties:

    - Statement: A [statement][pycfmodel.model.resources.properties.statement.Statement] object.
    - Version
    """

    class Config(Property.Config):
        extra = Extra.allow

    Statement: Resolvable[Union[Statement, List[Resolvable[Statement]]]]
    Version: Optional[ResolvableDate] = None

    def _statement_as_list(self) -> List[Statement]:
        if isinstance(self.Statement, Statement):
            return [self.Statement]
        return self.Statement

    def statements_with(self, pattern: Pattern) -> List[Statement]:
        """
        Finds all statements which have at least one resource with the pattern.

        Arguments:
            pattern: Pattern to match.

        Returns:
            List of [statements][pycfmodel.model.resources.properties.statement.Statement].
        """
        return [statement for statement in self._statement_as_list() if statement.resources_with(pattern)]

    def allowed_actions_with(self, pattern: Pattern) -> List[Statement]:
        """
        Finds all statements which have at least one action with the pattern.

        Arguments:
            pattern: Pattern to match.

        Returns:
            List of [statements][pycfmodel.model.resources.properties.statement.Statement].
        """
        return [
            statement
            for statement in self._statement_as_list()
            if statement.actions_with(pattern) and statement.Effect == "Allow"
        ]

    def allowed_principals_with(self, pattern: Pattern) -> List[str]:
        """
        Finds all allowed principals which match the pattern.

        Arguments:
            pattern: Pattern to match.

        Returns:
            List of principals.
        """
        principals = set()
        for statement in self._statement_as_list():
            if statement.Effect == "Allow":
                principals.update(statement.principals_with(pattern))
        return list(principals)

    def non_whitelisted_allowed_principals(self, whitelist: List[str]) -> List[str]:
        """
        Find non whitelisted allowed principals.

        Arguments:
            whitelist: List of whitelisted principals.

        Returns:
            List of principals.
        """
        principals = set()
        for statement in self._statement_as_list():
            if statement.Effect == "Allow":
                principals.update(statement.non_whitelisted_principals(whitelist))
        return list(principals)

    def get_iam_actions(self, difference=False) -> List[str]:
        """
        Find all IAM Actions which are specified in statements.

        Arguments:
            difference: when True, the behaviour changes to find the difference between all IAM Actions and those
            specified in the statements of the policy. Default = False.

        Returns:
            List of matching actions.
        """
        actions = set()
        for statement in self._statement_as_list():
            for action in statement.get_expanded_action_list():
                if action.startswith("iam:"):
                    actions.add(action)

        if difference:
            return sorted(
                set([action for action in CLOUDFORMATION_ACTIONS if action.lower().startswith("iam:")]).difference(
                    actions
                )
            )

        return sorted(actions)

    def get_allowed_actions(self) -> List[str]:
        """
        Find all allowed Actions which are specified in statements.

        Returns:
            List of matching actions.
        """
        actions = set()
        for statement in self._statement_as_list():
            if statement.Effect.lower() == "allow":
                actions.update(statement.get_expanded_action_list())
        return sorted(actions)



from typing import Any, ClassVar, List, Optional

from pydantic import PositiveInt

from pycfmodel.model.base import CustomModel


class Parameter(CustomModel):
    """
    CloudFormation Parameter object representation
    """

    NO_ECHO_NO_DEFAULT: ClassVar[str] = "NO_ECHO_NO_DEFAULT"
    NO_ECHO_WITH_DEFAULT: ClassVar[str] = "NO_ECHO_WITH_DEFAULT"
    NO_ECHO_WITH_VALUE: ClassVar[str] = "NO_ECHO_WITH_VALUE"
    AllowedPattern: Optional[str] = None
    AllowedValues: Optional[List] = None
    ConstraintDescription: Optional[str] = None
    Default: Optional[Any] = None
    Description: Optional[str] = None
    MaxLength: Optional[PositiveInt] = None
    MaxValue: Optional[PositiveInt] = None
    MinLength: Optional[int] = None
    MinValue: Optional[int] = None
    NoEcho: Optional[bool] = None
    Type: str

    def get_ref_value(self, provided_value=None) -> Optional[str]:
        """
        Calculates the parameter value to be used in the template.

        - If `NoEcho` property is set, it uses a constant value.
        - If it is a list of numbers or a comma delimited list, returns the string version of each element in a list.
        - Returns None if `provided_value` and `Default` are `None`.

        Arguments:
            provided_value: Value injected in the template

        Returns:
            The computed value.
        """
        value = provided_value if provided_value is not None else self.Default
        if self.NoEcho:
            if provided_value is not None:
                return self.NO_ECHO_WITH_VALUE
            elif self.Default:
                return self.NO_ECHO_WITH_DEFAULT
            else:
                return self.NO_ECHO_NO_DEFAULT

        elif self.Type in ["List<Number>", "CommaDelimitedList"]:
            return value.split(",")

        return value if value is None else str(value)



import pytest

from pycfmodel.model.resources.sns_topic_policy import SNSTopicPolicy
from pycfmodel.model.utils import OptionallyNamedPolicyDocument


@pytest.fixture()
def sns_topic_policy():
    return SNSTopicPolicy(
        **{
            "Type": "AWS::SNS::TopicPolicy",
            "Properties": {
                "PolicyDocument": {
                    "Id": "MyTopicPolicy",
                    "Version": "2012-10-17",
                    "Statement": {
                        "Sid": "My-statement-id",
                        "Effect": "Allow",
                        "Principal": "*",
                        "Action": "sns:Publish",
                        "Resource": "*",
                    },
                },
                "Topics": [{"Ref": "MySNSTopic"}],
            },
        }
    )


def test_sns_topic_policy(sns_topic_policy):
    assert len(sns_topic_policy.Properties.Topics) == 1
    assert sns_topic_policy.Properties.PolicyDocument.Statement.Effect == "Allow"


def test_sns_policy_documents(sns_topic_policy):
    assert sns_topic_policy.policy_documents == [
        OptionallyNamedPolicyDocument(name=None, policy_document=sns_topic_policy.Properties.PolicyDocument)
    ]
